"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import torch

from nataili.util.logger import logger
from nataili.util.voodoo import initialise_voodoo


class ModelManager:
    """
    Horde-specific SuperModelManager class.
    Note: This class is only intended for Stable Horde worker.
    New development should use the individual model managers instead.
    """

    def __init__(
        self,
        aitemplate: bool = False,
        blip: bool = False,
        clip: bool = False,
        compvis: bool = False,
        diffusers: bool = False,
        esrgan: bool = False,
        gfpgan: bool = False,
        safety_checker: bool = False,
        codeformer: bool = False,
        controlnet: bool = False,
    ):
        initialise_voodoo()
        if aitemplate:
            from nataili.model_manager.aitemplate import AITemplateModelManager

            self.aitemplate = AITemplateModelManager()
        else:
            self.aitemplate = None
        if blip:
            from nataili.model_manager.blip import BlipModelManager

            self.blip = BlipModelManager()
        else:
            self.blip = None
        if clip:
            from nataili.model_manager.clip import ClipModelManager

            self.clip = ClipModelManager()
        else:
            self.clip = None
        if compvis:
            from nataili.model_manager.compvis import CompVisModelManager

            self.compvis = CompVisModelManager()
        else:
            self.compvis = None
        if diffusers:
            from nataili.model_manager.diffusers import DiffusersModelManager

            self.diffusers = DiffusersModelManager()
        else:
            self.diffusers = None
        if esrgan:
            from nataili.model_manager.esrgan import EsrganModelManager

            self.esrgan = EsrganModelManager()
        else:
            self.esrgan = None
        if gfpgan:
            from nataili.model_manager.gfpgan import GfpganModelManager

            self.gfpgan = GfpganModelManager()
        else:
            self.gfpgan = None
        if safety_checker:
            from nataili.model_manager.safety_checker import SafetyCheckerModelManager

            self.safety_checker = SafetyCheckerModelManager()
        else:
            self.safety_checker = None
        if codeformer:
            from nataili.model_manager.codeformer import CodeFormerModelManager

            self.codeformer = CodeFormerModelManager()
        else:
            self.codeformer = None
        if controlnet:
            from nataili.model_manager.controlnet import ControlNetModelManager

            self.controlnet = ControlNetModelManager()
        else:
            self.controlnet = None
        self.cuda_available = torch.cuda.is_available()
        self.models = {}
        self.available_models = []
        self.loaded_models = {}
        self.init()

    def init(self):
        """
        Initialize SuperModelManager's models and available_models from
        the models and available_models of the model types.
        Individual model types are already initialized in their own init() functions
        which are called when the individual model manager is created in __init__.
        """
        model_types = [
            self.aitemplate,
            self.blip,
            self.clip,
            self.compvis,
            self.diffusers,
            self.esrgan,
            self.gfpgan,
            self.safety_checker,
            self.codeformer,
            self.controlnet,
        ]
        # reset available models
        self.available_models = []
        for model_type in model_types:
            if model_type is not None:
                self.models.update(model_type.models)
                self.available_models.extend(model_type.available_models)

    def reload_database(self):
        """
        Horde-specific function to reload the database of available models.
        Note: It is not appropriate to place `model_type.init()` in `init()`
        because individual model types are already initialized after being created
        i.e. if `model_type.init()` is placed in `self.init()`, the database will be
        loaded twice.
        """
        model_types = [
            self.aitemplate,
            self.blip,
            self.clip,
            self.compvis,
            self.diffusers,
            self.esrgan,
            self.gfpgan,
            self.safety_checker,
            self.codeformer,
            self.controlnet,
        ]
        self.available_models = []  # reset available models
        for model_type in model_types:
            if model_type is not None:
                model_type.init()
                self.models.update(model_type.models)
                self.available_models.extend(model_type.available_models)

    def download_model(self, model_name):
        if self.aitemplate is not None and model_name in self.aitemplate.models:
            return self.aitemplate.download_model(model_name)
        if self.blip is not None and model_name in self.blip.models:
            return self.blip.download_model(model_name)
        if self.clip is not None and model_name in self.clip.models:
            return self.clip.download_model(model_name)
        if self.codeformer is not None and model_name in self.codeformer.models:
            return self.codeformer.download_model(model_name)
        if self.compvis is not None and model_name in self.compvis.models:
            return self.compvis.download_model(model_name)
        if self.diffusers is not None and model_name in self.diffusers.models:
            return self.diffusers.download_model(model_name)
        if self.esrgan is not None and model_name in self.esrgan.models:
            return self.esrgan.download_model(model_name)
        if self.gfpgan is not None and model_name in self.gfpgan.models:
            return self.gfpgan.download_model(model_name)
        if self.safety_checker is not None and model_name in self.safety_checker.models:
            return self.safety_checker.download_model(model_name)
        if self.controlnet is not None and model_name in self.controlnet.models:
            return self.controlnet.download_model(model_name)

    def download_all(self):
        if self.aitemplate is not None:
            self.aitemplate.download_ait()
        if self.blip is not None:
            self.blip.download_all_models()
        if self.clip is not None:
            self.clip.download_all_models()
        if self.codeformer is not None:
            self.codeformer.download_all_models()
        if self.compvis is not None:
            self.compvis.download_all_models()
        if self.diffusers is not None:
            self.diffusers.download_all_models()
        if self.esrgan is not None:
            self.esrgan.download_all_models()
        if self.gfpgan is not None:
            self.gfpgan.download_all_models()
        if self.safety_checker is not None:
            self.safety_checker.download_all_models()
        if self.controlnet is not None:
            self.controlnet.download_all_models()

    def validate_model(self, model_name, skip_checksum=False):
        if self.blip is not None and model_name in self.blip.models:
            return self.blip.validate_model(model_name, skip_checksum)
        if self.clip is not None and model_name in self.clip.models:
            return self.clip.validate_model(model_name, skip_checksum)
        if self.codeformer is not None and model_name in self.codeformer.models:
            return self.codeformer.validate_model(model_name, skip_checksum)
        if self.compvis is not None and model_name in self.compvis.models:
            return self.compvis.validate_model(model_name, skip_checksum)
        if self.diffusers is not None and model_name in self.diffusers.models:
            return self.diffusers.validate_model(model_name, skip_checksum)
        if self.esrgan is not None and model_name in self.esrgan.models:
            return self.esrgan.validate_model(model_name, skip_checksum)
        if self.gfpgan is not None and model_name in self.gfpgan.models:
            return self.gfpgan.validate_model(model_name, skip_checksum)
        if self.safety_checker is not None and model_name in self.safety_checker.models:
            return self.safety_checker.validate_model(model_name, skip_checksum)
        if self.controlnet is not None and model_name in self.controlnet.models:
            return self.controlnet.validate_model(model_name, skip_checksum)

    def taint_models(self, models):
        if self.aitemplate is not None and any(model in self.aitemplate.models for model in models):
            self.aitemplate.taint_models(models)
        if self.blip is not None and any(model in self.blip.models for model in models):
            self.blip.taint_models(models)
        if self.clip is not None and any(model in self.clip.models for model in models):
            self.clip.taint_models(models)
        if self.codeformer is not None and any(model in self.codeformer.models for model in models):
            self.codeformer.taint_models(models)
        if self.compvis is not None and any(model in self.compvis.models for model in models):
            self.compvis.taint_models(models)
        if self.diffusers is not None and any(model in self.diffusers.models for model in models):
            self.diffusers.taint_models(models)
        if self.esrgan is not None and any(model in self.esrgan.models for model in models):
            self.esrgan.taint_models(models)
        if self.gfpgan is not None and any(model in self.gfpgan.models for model in models):
            self.gfpgan.taint_models(models)
        if self.safety_checker is not None and any(model in self.safety_checker.models for model in models):
            self.safety_checker.taint_models(models)
        if self.controlnet is not None and any(model in self.controlnet.models for model in models):
            self.controlnet.taint_models(models)

    def unload_model(self, model_name):
        if self.aitemplate is not None and model_name in self.aitemplate.models:
            self.aitemplate.unload_model(model_name)
        if self.blip is not None and model_name in self.blip.models:
            self.blip.unload_model(model_name)
            del self.loaded_models[model_name]
        if self.clip is not None and model_name in self.clip.models:
            self.clip.unload_model(model_name)
            del self.loaded_models[model_name]
        if self.codeformer is not None and model_name in self.codeformer.models:
            self.codeformer.unload_model(model_name)
            del self.loaded_models[model_name]
        if self.compvis is not None and model_name in self.compvis.models:
            self.compvis.unload_model(model_name)
            del self.loaded_models[model_name]
        if self.diffusers is not None and model_name in self.diffusers.models:
            self.diffusers.unload_model(model_name)
            del self.loaded_models[model_name]
        if self.esrgan is not None and model_name in self.esrgan.models:
            self.esrgan.unload_model(model_name)
            del self.loaded_models[model_name]
        if self.gfpgan is not None and model_name in self.gfpgan.models:
            self.gfpgan.unload_model(model_name)
            del self.loaded_models[model_name]
        if self.safety_checker is not None and model_name in self.safety_checker.models:
            self.safety_checker.unload_model(model_name)
            del self.loaded_models[model_name]
        if self.controlnet is not None and model_name in self.controlnet.models:
            self.controlnet.unload_model(model_name)
            del self.loaded_models[model_name]
        # Also remove from super() model list
        if model_name in self.loaded_models:
            del self.loaded_models[model_name]

    def get_loaded_models_names(self, string=False):
        """
        :param string: If True, returns concatenated string of model names
        Returns a list of the loaded model names
        """
        if string:
            return ", ".join(self.loaded_models.keys())
        return list(self.loaded_models.keys())

    def get_available_models_by_types(self, model_types=None):
        if not model_types:
            model_types = ["ckpt", "diffusers"]
        models_available = []
        for model_type in model_types:
            if model_type == "ckpt":
                if self.compvis is not None:
                    for model in self.compvis.models:
                        # We don't want to check the .yaml file as those exist in this repo instead
                        model_files = [
                            filename
                            for filename in self.compvis.get_model_files(model)
                            if not filename["path"].endswith(".yaml")
                        ]
                        if self.compvis.check_available(model_files):
                            models_available.append(model)
            if model_type == "diffusers":
                if self.diffusers is not None:
                    for model in self.diffusers.models:
                        if self.diffusers.check_available(self.diffusers.get_model_files(model)):
                            models_available.append(model)
        return models_available

    def count_available_models_by_types(self, model_types=None):
        return len(self.get_available_models_by_types(model_types))

    def get_available_models(self):
        """
        Returns the available models
        """
        return self.available_models

    def load(
        self,
        model_name,
        half_precision=True,
        gpu_id=0,
        cpu_only=False,
        voodoo=False,
    ):
        """
        model_name: str. Name of the model to load. See available_models for a list of available models.
        half_precision: bool. If True, the model will be loaded in half precision.
        gpu_id: int. The id of the gpu to use. If the gpu is not available, the model will be loaded on the cpu.
        cpu_only: bool. If True, the model will be loaded on the cpu. If True, half_precision will be set to False.
        voodoo: bool. (compvis only) Voodoo ray.
        """
        if not self.cuda_available:
            cpu_only = True
        if self.aitemplate is not None and model_name in self.aitemplate.models:
            return self.aitemplate.load(model_name, gpu_id)
        if self.blip is not None and model_name in self.blip.models:
            success = self.blip.load(
                model_name=model_name, half_precision=half_precision, gpu_id=gpu_id, cpu_only=cpu_only
            )
            if success:
                self.loaded_models.update({model_name: self.blip.loaded_models[model_name]})
            return success
        if self.clip is not None and model_name in self.clip.models:
            success = self.clip.load(
                model_name=model_name, half_precision=half_precision, gpu_id=gpu_id, cpu_only=cpu_only
            )
            if success:
                self.loaded_models.update({model_name: self.clip.loaded_models[model_name]})
            return success
        if self.codeformer is not None and model_name in self.codeformer.models:
            success = self.codeformer.load(
                model_name=model_name, half_precision=half_precision, gpu_id=gpu_id, cpu_only=cpu_only
            )
            if success:
                self.loaded_models.update({model_name: self.codeformer.loaded_models[model_name]})
            return success
        if self.compvis is not None and model_name in self.compvis.models:
            success = self.compvis.load(
                model_name=model_name, half_precision=half_precision, gpu_id=gpu_id, cpu_only=cpu_only, voodoo=voodoo
            )
            if success:
                self.loaded_models.update({model_name: self.compvis.loaded_models[model_name]})
            return success
        if self.diffusers is not None and model_name in self.diffusers.models:
            success = self.diffusers.load(
                model_name=model_name, half_precision=half_precision, gpu_id=gpu_id, cpu_only=cpu_only, voodoo=voodoo
            )
            if success:
                self.loaded_models.update({model_name: self.diffusers.loaded_models[model_name]})
            return success
        if self.esrgan is not None and model_name in self.esrgan.models:
            success = self.esrgan.load(
                model_name=model_name, half_precision=half_precision, gpu_id=gpu_id, cpu_only=cpu_only
            )
            if success:
                self.loaded_models.update({model_name: self.esrgan.loaded_models[model_name]})
            return success
        if self.gfpgan is not None and model_name in self.gfpgan.models:
            success = self.gfpgan.load(model_name=model_name, gpu_id=gpu_id, cpu_only=cpu_only)
            if success:
                self.loaded_models.update({model_name: self.gfpgan.loaded_models[model_name]})
            return success
        if self.safety_checker is not None and model_name in self.safety_checker.models:
            success = self.safety_checker.load(
                model_name=model_name, half_precision=half_precision, gpu_id=gpu_id, cpu_only=True  # for the horde
            )
            if success:
                self.loaded_models.update({model_name: self.safety_checker.loaded_models[model_name]})
            return success
        logger.error(f"{model_name} not found")
        return False



"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os
import random
from typing import Literal, Union

import numpy as np
import torch
import torch.nn as nn
import wandb
from torch.utils.data import DataLoader, TensorDataset
from tqdm import tqdm

from nataili.clip.predictor.mlp import H14_MLP, H14_MLP_2, MLP
from nataili.util.logger import logger


class TrainPredictor:
    def __init__(
        self,
        x: str,
        y: str,
        batch_size: int = 256,
        vit_model: Literal["ViT-L/14", "ViT-H-14"] = "ViT-L/14",
        output_directory: str = "output",
        save_name: str = "model.pth",
        lr: float = 1e-3,
        optimizer: Literal["adam", "adamw"] = "adam",
        validation_percentage: float = 0.05,
        number_of_train_epochs: int = 100,
        number_of_validation_samples: int = 10,
        device: Literal["cpu", "cuda"] = "cuda",
        gpu_id: int = 0,
        project_name: str = "a e s t h e t i c s",
        wandb_entity: str = "hlky",
        mlp_model: Union[MLP, H14_MLP, H14_MLP_2] = MLP,
    ):
        self.x = None
        self.y = None
        self.val_x = None
        self.val_y = None
        self.batch_size = batch_size
        self.output_directory = output_directory
        self.save_name = f"{output_directory}/{save_name}"
        self.lr = lr
        self.optimizer = optimizer
        self.validation_percentage = validation_percentage
        self.number_of_train_epochs = number_of_train_epochs
        self.number_of_validation_samples = number_of_validation_samples
        self.mlp_model = mlp_model
        if device == "cuda":
            self.device = torch.device(f"cuda:{gpu_id}")
        else:
            self.device = torch.device("cpu")
        if vit_model == "ViT-L/14":
            self.model = self.mlp_model(input_size=768).to(self.device)
        elif vit_model == "ViT-H-14":
            self.model = self.mlp_model(input_size=1024).to(self.device)
        else:
            raise NotImplementedError
        self.best_loss = 999
        self.best_acc = 0
        os.makedirs(self.output_directory, exist_ok=True)
        if os.path.exists(self.save_name):
            raise FileExistsError(f"{self.save_name} already exists")
        self.load_dataset(x, y)
        assert len(self.x) == len(self.y)
        self.load_val_dataset()
        assert len(self.val_x) == len(self.val_y)
        self.train_x_tensor = torch.Tensor(self.x).to(self.device)
        self.train_y_tensor = torch.Tensor(self.y).to(self.device)
        self.val_x_tensor = torch.Tensor(self.val_x).to(self.device)
        self.val_y_tensor = torch.Tensor(self.val_y).to(self.device)
        self.train_dataset = TensorDataset(self.train_x_tensor, self.train_y_tensor)
        self.val_dataset = TensorDataset(self.val_x_tensor, self.val_y_tensor)
        self.train_loader = DataLoader(self.train_dataset, shuffle=True, batch_size=self.batch_size)
        self.val_loader = DataLoader(self.val_dataset, batch_size=self.batch_size)
        self.criterion = nn.MSELoss()
        self.current_epoch = 0
        self.config = {
            "vit": vit_model,
            "images_count": len(self.x),
            "epochs": self.number_of_train_epochs,
            "validation_samples": self.number_of_validation_samples,
            "lr": self.lr,
            "optimizer": self.optimizer,
            "save_name": self.save_name,
        }
        wandb.init(project=project_name, entity=wandb_entity, config=self.config, reinit=True)

    def load_dataset(self, x, y):
        try:
            self.x = np.load(x)
        except Exception as e:
            logger.error(f"Could not load {self.x} - {e}")
            exit(1)
        try:
            self.y = np.load(y)
        except Exception as e:
            logger.error(f"Could not load {self.y} - {e}")
            exit(1)

    def load_val_dataset(self):
        border = int(len(self.x) * self.validation_percentage)
        self.x = self.x[border:]
        self.y = self.y[border:]
        self.val_x = self.x[:border]
        self.val_y = self.y[:border]

    def train_one_epoch(self):
        self.model.train()
        for x, y in self.train_loader:
            self.model.optimizer.zero_grad()
            y_pred = self.model(x)
            loss = self.criterion(y_pred, y)
            loss.backward()
            self.model.optimizer.step()
        return loss.item()

    def validate(self):
        self.model.eval()
        with torch.no_grad():
            for x, y in self.val_loader:
                y_pred = self.model(x)
                loss = self.criterion(y_pred, y)
        return loss.item()

    def val_accuracy(self, tolerance=0.5, low=2, high=6):
        correct = 0
        total = 0
        self.model.eval()
        with torch.no_grad():
            for x, y in self.val_loader:
                y_pred = self.model(x)
                for i in range(len(y_pred)):
                    total += 1
                    if (
                        abs(y_pred[i].item() - y[i].item()) <= tolerance  # predicted is within tolerance of actual
                        or (
                            y[i].item() >= high and y_pred[i].item() >= y[i].item()
                        )  # predicted is higher than actual and actual is higher than high
                        or (
                            y[i].item() <= low and y_pred[i].item() <= y[i].item()
                        )  # predicted is lower than actual and actual is lower than low
                    ):
                        correct += 1
        return correct / total

    def __call__(self):
        for epoch in range(self.number_of_train_epochs):
            self.current_epoch = epoch
            logger.info(f"Epoch: {self.current_epoch}")
            train_loss = self.train_one_epoch()
            val_loss = self.validate()
            val_acc = self.val_accuracy()
            wandb.log(
                {
                    "train_loss": train_loss,
                    "val_loss": val_loss,
                    "val_accuracy": val_acc,
                }
            )
            logger.info(f"Val loss: {val_loss} - Val accuracy: {val_acc}")
            logger.info(f"Train loss: {train_loss}")
            if val_acc > self.best_acc:
                self.best_acc = val_acc
                torch.save(self.model.state_dict(), self.save_name)
                logger.info(f"Saved model with val_acc {round(val_acc, 4)} at epoch {epoch}")
        self.model.load_state_dict(torch.load(self.save_name))
        return True



import torch
import torch.nn.functional as F
from models.med import BertConfig, BertModel
from torch import nn
from transformers import BertTokenizer

from models.blip import create_vit, init_tokenizer, load_checkpoint


class BLIP_ITM(nn.Module):
    def __init__(self,                 
                 med_config = 'configs/med_config.json',  
                 image_size = 384,
                 vit = 'base',
                 vit_grad_ckpt = False,
                 vit_ckpt_layer = 0,                      
                 embed_dim = 256,     
                 ):
        """
        Args:
            med_config (str): path for the mixture of encoder-decoder model's configuration file
            image_size (int): input image size
            vit (str): model size of vision transformer
        """               
        super().__init__()
        
        self.visual_encoder, vision_width = create_vit(vit,image_size, vit_grad_ckpt, vit_ckpt_layer)
        self.tokenizer = init_tokenizer()   
        med_config = BertConfig.from_json_file(med_config)
        med_config.encoder_width = vision_width
        self.text_encoder = BertModel(config=med_config, add_pooling_layer=False)          

        text_width = self.text_encoder.config.hidden_size
        
        self.vision_proj = nn.Linear(vision_width, embed_dim)
        self.text_proj = nn.Linear(text_width, embed_dim)

        self.itm_head = nn.Linear(text_width, 2) 
        
        
    def forward(self, image, caption, match_head='itm'):

        image_embeds = self.visual_encoder(image) 
        image_atts = torch.ones(image_embeds.size()[:-1],dtype=torch.long).to(image.device)        
      
        text = self.tokenizer(caption, padding='max_length', truncation=True, max_length=35, 
                              return_tensors="pt").to(image.device) 

                 
        if match_head=='itm':
            output = self.text_encoder(text.input_ids,
                                       attention_mask = text.attention_mask,
                                       encoder_hidden_states = image_embeds,
                                       encoder_attention_mask = image_atts,      
                                       return_dict = True,
                                      )
            itm_output = self.itm_head(output.last_hidden_state[:,0,:])     
            return itm_output
            
        elif match_head=='itc':
            text_output = self.text_encoder(text.input_ids, attention_mask = text.attention_mask,                      
                                            return_dict = True, mode = 'text')                     
            image_feat = F.normalize(self.vision_proj(image_embeds[:,0,:]),dim=-1)   
            text_feat = F.normalize(self.text_proj(text_output.last_hidden_state[:,0,:]),dim=-1)    
            
            sim = image_feat @ text_feat.t()
            return sim
        
        
def blip_itm(pretrained='',**kwargs):
    model = BLIP_ITM(**kwargs)
    if pretrained:
        model,msg = load_checkpoint(model,pretrained)
        assert(len(msg.missing_keys)==0)
    return model         
            


from typing import Any, Callable, Dict, List, Union

import gradio as gr

from nataili.model_manager.blip import BlipModelManager
from nataili.model_manager.clip import ClipModelManager
from nataili.model_manager.codeformer import CodeFormerModelManager
from nataili.model_manager.compvis import CompVisModelManager
from nataili.model_manager.diffusers import DiffusersModelManager
from nataili.model_manager.esrgan import EsrganModelManager
from nataili.model_manager.gfpgan import GfpganModelManager
from nataili.model_manager.safety_checker import SafetyCheckerModelManager


class GradioUI:
    def __init__(
        self,
        inputs,
        outputs,
        model_manager_class: Union[
            BlipModelManager,
            ClipModelManager,
            CodeFormerModelManager,
            CompVisModelManager,
            DiffusersModelManager,
            EsrganModelManager,
            GfpganModelManager,
            SafetyCheckerModelManager,
        ] = None,
        model_name: str = "",
        title: str = "",
        description: str = "",
        thumbnail: str = "",
    ):
        self.model_manager_class = model_manager_class
        self.model_name = model_name
        self.inputs = inputs
        self.outputs = outputs
        self.title = title
        self.description = description
        self.thumbnail = thumbnail
        self.model_manager = None

    def launch(self, predict_function: Callable):
        if self.model_manager_class is not None and self.model_name != "":
            self.model_manager = self.model_manager_class()
            self.model_manager.load(self.model_name)

        interface = gr.Interface(
            predict_function,
            self.inputs,
            self.outputs,
            title=self.title,
            description=self.description,
            thumbnail=self.thumbnail,
            analytics_enabled=False,
            allow_flagging="never",
        )

        interface.launch(show_error=True, quiet=False, debug=True)



# dataset settings
dataset_type = 'CityscapesDataset'
data_root = 'data/cityscapes/'
img_norm_cfg = dict(
    mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)
crop_size = (512, 1024)
train_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations'),
    dict(type='Resize', img_scale=(2048, 1024), ratio_range=(0.5, 2.0)),
    dict(type='RandomCrop', crop_size=crop_size, cat_max_ratio=0.75),
    dict(type='RandomFlip', prob=0.5),
    dict(type='PhotoMetricDistortion'),
    dict(type='Normalize', **img_norm_cfg),
    dict(type='Pad', size=crop_size, pad_val=0, seg_pad_val=255),
    dict(type='DefaultFormatBundle'),
    dict(type='Collect', keys=['img', 'gt_semantic_seg']),
]
test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(
        type='MultiScaleFlipAug',
        img_scale=(2048, 1024),
        # img_ratios=[0.5, 0.75, 1.0, 1.25, 1.5, 1.75],
        flip=False,
        transforms=[
            dict(type='Resize', keep_ratio=True),
            dict(type='RandomFlip'),
            dict(type='Normalize', **img_norm_cfg),
            dict(type='ImageToTensor', keys=['img']),
            dict(type='Collect', keys=['img']),
        ])
]
data = dict(
    samples_per_gpu=2,
    workers_per_gpu=2,
    train=dict(
        type=dataset_type,
        data_root=data_root,
        img_dir='leftImg8bit/train',
        ann_dir='gtFine/train',
        pipeline=train_pipeline),
    val=dict(
        type=dataset_type,
        data_root=data_root,
        img_dir='leftImg8bit/val',
        ann_dir='gtFine/val',
        pipeline=test_pipeline),
    test=dict(
        type=dataset_type,
        data_root=data_root,
        img_dir='leftImg8bit/val',
        ann_dir='gtFine/val',
        pipeline=test_pipeline))



"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from typing import Dict, List, Tuple, Union

import cv2
import einops
import numpy as np
import torch
from PIL import Image

from annotator.canny import CannyDetector
from annotator.hed import HEDdetector, nms
from annotator.midas import MidasDetector
from annotator.mlsd import MLSDdetector
from annotator.openpose import OpenposeDetector
from annotator.uniformer import UniformerDetector
from annotator.util import HWC3, resize_image


class Annotation:
    def __init__(self, image_is_control: bool = False):
        self.image_is_control = image_is_control
        self.model = None


class Canny(Annotation):
    def __init__(self, image_is_control: bool = False):
        super().__init__(image_is_control)
        self.model = CannyDetector()

    def __call__(
        self,
        image: Image.Image,
        low_threshold: int = 100,
        high_threshold: int = 200,
        resolution: int = 512,
        num_samples: int = 1,
    ):
        image: np.ndarray = np.asarray(image)
        if self.image_is_control:
            detected_map = HWC3(image)
            _image = resize_image(image, resolution)
            H, W, C = _image.shape
        else:
            image = resize_image(HWC3(image), resolution)
            H, W, C = image.shape
            detected_map = self.model(image, low_threshold, high_threshold)
            detected_map = HWC3(detected_map)

        control = torch.from_numpy(detected_map.copy()).float() / 255.0
        control = torch.stack([control for _ in range(num_samples)], dim=0)
        control = einops.rearrange(control, "b h w c -> b c h w").clone()
        return {"control": control, "detected_map": [255 - detected_map], "shape": (H, W, C)}


class HED(Annotation):
    def __init__(self, image_is_control: bool = False):
        super().__init__(image_is_control)
        self.model = HEDdetector()

    def __call__(self, image: Image.Image, resolution: int = 512, detect_resolution: int = 512, num_samples: int = 1):
        image: np.ndarray = np.asarray(image)
        if self.image_is_control:
            detected_map = HWC3(image)
            _image = resize_image(image, resolution)
            H, W, C = _image.shape
        else:
            image = HWC3(image)
            detected_map = self.model(resize_image(image, detect_resolution))
            image = resize_image(image, resolution)
            H, W, C = image.shape
            detected_map = HWC3(detected_map)

        detected_map = cv2.resize(detected_map, (W, H), interpolation=cv2.INTER_LINEAR)
        control = torch.from_numpy(detected_map.copy()).float() / 255.0
        control = torch.stack([control for _ in range(num_samples)], dim=0)
        control = einops.rearrange(control, "b h w c -> b c h w").clone()
        return {"control": control, "detected_map": [detected_map], "shape": (H, W, C)}


class FakeScribbles(Annotation):
    def __init__(self, image_is_control: bool = False):
        super().__init__(image_is_control)
        self.model = HEDdetector()

    def __call__(self, image: Image.Image, resolution: int = 512, detect_resolution: int = 512, num_samples: int = 1):
        image: np.ndarray = np.asarray(image)
        if self.image_is_control:
            detected_map = HWC3(image)
            _image = resize_image(image, resolution)
            H, W, C = _image.shape
        else:
            image = HWC3(image)
            detected_map = self.model(resize_image(image, detect_resolution))
            image = resize_image(image, resolution)
            H, W, C = image.shape
            detected_map = HWC3(detected_map)

        detected_map = cv2.resize(detected_map, (W, H), interpolation=cv2.INTER_LINEAR)
        detected_map = nms(detected_map, 127, 3.0)
        detected_map = cv2.GaussianBlur(detected_map, (0, 0), 3.0)
        detected_map[detected_map > 4] = 255
        detected_map[detected_map < 255] = 0

        control = torch.from_numpy(detected_map.copy()).float() / 255.0
        control = torch.stack([control for _ in range(num_samples)], dim=0)
        control = einops.rearrange(control, "b h w c -> b c h w").clone()
        return {"control": control, "detected_map": [255 - detected_map], "shape": (H, W, C)}


class Hough(Annotation):
    def __init__(self, image_is_control: bool = False):
        super().__init__(image_is_control)
        self.model = MLSDdetector()

    def __call__(
        self,
        image: Image.Image,
        value_threshold: float = 0.1,
        distance_threshold: float = 0.1,
        resolution: int = 512,
        detect_resolution: int = 512,
        num_samples: int = 1,
    ):
        image: np.ndarray = np.asarray(image)
        if self.image_is_control:
            detected_map = HWC3(image)
            _image = resize_image(image, resolution)
            H, W, C = _image.shape
        else:
            image = HWC3(image)
            detected_map = self.model(resize_image(image, detect_resolution), value_threshold, distance_threshold)
            image = resize_image(image, resolution)
            H, W, C = image.shape
            detected_map = HWC3(detected_map)

        detected_map = cv2.resize(detected_map, (W, H), interpolation=cv2.INTER_NEAREST)
        control = torch.from_numpy(detected_map.copy()).float() / 255.0
        control = torch.stack([control for _ in range(num_samples)], dim=0)
        control = einops.rearrange(control, "b h w c -> b c h w").clone()
        return {
            "control": control,
            "detected_map": [255 - cv2.dilate(detected_map, np.ones(shape=(3, 3), dtype=np.uint8), iterations=1)],
            "shape": (H, W, C),
        }


class Depth(Annotation):
    def __init__(self, image_is_control: bool = False):
        super().__init__(image_is_control)
        self.model = MidasDetector()

    def __call__(self, image: Image.Image, resolution: int = 512, depth_resolution: int = 384, num_samples: int = 1):
        image: np.ndarray = np.asarray(image)
        if self.image_is_control:
            detected_map = HWC3(image)
            """
            the provided depth map might be depth_resolution already so we need to resize it to resolution
            """
            _image = resize_image(image, resolution)
            H, W, C = _image.shape
        else:
            image = HWC3(image)
            detected_map, _ = self.model(resize_image(image, depth_resolution))
            image = resize_image(image, resolution)
            H, W, C = image.shape
            detected_map = HWC3(detected_map)

        detected_map = cv2.resize(detected_map, (W, H), interpolation=cv2.INTER_LINEAR)
        control = torch.from_numpy(detected_map.copy()).float() / 255.0
        control = torch.stack([control for _ in range(num_samples)], dim=0)
        control = einops.rearrange(control, "b h w c -> b c h w").clone()
        return {"control": control, "detected_map": [detected_map], "shape": (H, W, C)}


class Normal(Annotation):
    def __init__(self, image_is_control: bool = False):
        super().__init__(image_is_control)
        self.model = MidasDetector()

    def __call__(
        self,
        image: Image.Image,
        background_threshold: float = 0.4,
        resolution: int = 512,
        depth_resolution: int = 384,
        num_samples: int = 1,
    ):
        image: np.ndarray = np.asarray(image)
        if self.image_is_control:
            detected_map = HWC3(image)
            """
            the provided depth map might be depth_resolution already so we need to resize it to resolution
            """
            _image = resize_image(image, resolution)
            H, W, C = _image.shape
        else:
            image = HWC3(image)
            _, detected_map = self.model(resize_image(image, depth_resolution), bg_th=background_threshold)
            image = resize_image(image, resolution)
            H, W, C = image.shape
            detected_map = HWC3(detected_map)

        detected_map = cv2.resize(detected_map, (W, H), interpolation=cv2.INTER_LINEAR)
        control = torch.from_numpy(detected_map[:, :, ::-1].copy()).float() / 255.0
        control = torch.stack([control for _ in range(num_samples)], dim=0)
        control = einops.rearrange(control, "b h w c -> b c h w").clone()
        return {"control": control, "detected_map": [detected_map], "shape": (H, W, C)}


class Openpose(Annotation):
    def __init__(self, image_is_control: bool = False):
        super().__init__(image_is_control)
        self.model = OpenposeDetector()

    def __call__(
        self,
        image: Image.Image,
        has_hand: bool = False,
        resolution: int = 512,
        detect_resolution: int = 512,
        num_samples: int = 1,
    ):
        image: np.ndarray = np.asarray(image)
        if self.image_is_control:
            detected_map = HWC3(image)
            _image = resize_image(image, resolution)
            H, W, C = _image.shape
        else:
            image = HWC3(image)
            detected_map, _ = self.model(resize_image(image, detect_resolution), has_hand)
            image = resize_image(image, resolution)
            detected_map = HWC3(detected_map)
            H, W, C = image.shape

        detected_map = cv2.resize(detected_map, (W, H), interpolation=cv2.INTER_NEAREST)
        control = torch.from_numpy(detected_map.copy()).float() / 255.0
        control = torch.stack([control for _ in range(num_samples)], dim=0)
        control = einops.rearrange(control, "b h w c -> b c h w").clone()
        return {"control": control, "detected_map": [detected_map], "shape": (H, W, C)}


class Seg(Annotation):
    def __init__(self, image_is_control: bool = False):
        super().__init__(image_is_control)
        self.model = UniformerDetector()

    def __call__(self, image: Image.Image, resolution: int = 512, detect_resolution: int = 512, num_samples: int = 1):
        image: np.ndarray = np.asarray(image)
        if self.image_is_control:
            detected_map = HWC3(image)
            _image = resize_image(image, resolution)
            H, W, C = _image.shape
        else:
            image = HWC3(image)
            detected_map = self.model(resize_image(image, detect_resolution))
            image = resize_image(image, resolution)
            H, W, C = image.shape

        detected_map = cv2.resize(detected_map, (W, H), interpolation=cv2.INTER_NEAREST)
        control = torch.from_numpy(detected_map.copy()).float() / 255.0
        control = torch.stack([control for _ in range(num_samples)], dim=0)
        control = einops.rearrange(control, "b h w c -> b c h w").clone()
        return {"control": control, "detected_map": [detected_map], "shape": (H, W, C)}


class Scribble(Annotation):
    def __init__(self, image_is_control: bool = False):
        super().__init__(image_is_control)
        self.model = None

    def __call__(self, image: Image.Image, resolution: int = 512, num_samples: int = 1):
        """
        image can't be used as control here
        """
        image: np.ndarray = np.asarray(image)
        image = resize_image(HWC3(image), resolution)
        H, W, C = image.shape
        detected_map = np.zeros_like(image, dtype=np.uint8)
        detected_map[np.min(image, axis=2) < 127] = 255

        control = torch.from_numpy(detected_map.copy()).float() / 255.0
        control = torch.stack([control for _ in range(num_samples)], dim=0)
        control = einops.rearrange(control, "b h w c -> b c h w").clone()
        return {"control": control, "detected_map": [255 - detected_map], "shape": (H, W, C)}



"""
Copyright [2022] Victor C Hall

Licensed under the GNU Affero General Public License;
You may not use this code except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/agpl-3.0.en.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging
import math
import random

import numpy
import torch
import torch.nn.functional as F
from torch.utils.data import Dataset
from torchvision import transforms
from transformers import CLIPTokenizer

from . import dl_singleton as dls
from .data_loader import DataLoaderMultiAspect as dlma
from .image_train_item import ImageTrainItem


class EveryDreamBatch(Dataset):
    """
    data_root: root path of all your training images, will be recursively searched for images
    repeats: how many times to repeat each image in the dataset
    flip_p: probability of flipping the image horizontally
    debug_level: 0=none, 1=print drops due to unfilled batches on aspect ratio buckets, 2=debug info per image, 3=save crops to disk for inspection
    batch_size: how many images to return in a batch
    conditional_dropout: probability of dropping the caption for a given image
    resolution: max resolution (relative to square)
    jitter: number of pixels to jitter the crop by, only for non-square images
    """

    def __init__(
        self,
        data_root,
        flip_p=0.0,
        debug_level=0,
        batch_size=1,
        conditional_dropout=0.02,
        resolution=512,
        crop_jitter=20,
        seed=555,
        tokenizer=None,
        log_folder=None,
        retain_contrast=False,
        write_schedule=False,
    ):
        self.data_root = data_root
        self.batch_size = batch_size
        self.debug_level = debug_level
        self.conditional_dropout = conditional_dropout
        self.crop_jitter = crop_jitter
        self.unloaded_to_idx = 0
        self.tokenizer = tokenizer
        self.log_folder = log_folder
        # print(f"tokenizer: {tokenizer}")
        self.max_token_length = self.tokenizer.model_max_length
        self.retain_contrast = retain_contrast
        self.write_schedule = write_schedule

        if seed == -1:
            seed = random.randint(0, 99999)

        if not dls.shared_dataloader:
            logging.info(" * Creating new dataloader singleton")
            dls.shared_dataloader = dlma(
                data_root=data_root,
                seed=seed,
                debug_level=debug_level,
                batch_size=self.batch_size,
                flip_p=flip_p,
                resolution=resolution,
                log_folder=self.log_folder,
            )

        self.image_train_items = dls.shared_dataloader.get_all_images()

        self.num_images = len(self.image_train_items)

        self._length = self.num_images

        logging.info(
            f" ** Trainer Set: {self._length / batch_size:.0f}, num_images: {self.num_images}, batch_size: {self.batch_size}"
        )
        if self.write_schedule:
            self.write_batch_schedule(0)

    def write_batch_schedule(self, epoch_n):
        with open(f"{self.log_folder}/ep{epoch_n}_batch_schedule.txt", "w", encoding="utf-8") as f:
            for i in range(len(self.image_train_items)):
                try:
                    f.write(
                        f"step:{int(i / self.batch_size):05}, wh:{self.image_train_items[i].target_wh}, r:{self.image_train_items[i].runt_size}, path:{self.image_train_items[i].pathname}\n"
                    )
                except Exception as e:
                    logging.error(
                        f" * Error writing to batch schedule for file path: {self.image_train_items[i].pathname}"
                    )

    def get_runts():
        return dls.shared_dataloader.runts

    def shuffle(self, epoch_n):
        if dls.shared_dataloader:
            dls.shared_dataloader.shuffle()
            self.image_train_items = dls.shared_dataloader.get_all_images()
        else:
            raise Exception("No dataloader singleton to shuffle")

        if self.write_schedule:
            self.write_batch_schedule(epoch_n)

    def __len__(self):
        return self._length

    def __getitem__(self, i):
        example = {}

        train_item = self.__get_image_for_trainer(self.image_train_items[i], self.debug_level)

        if self.retain_contrast:
            std_dev = 1.0
            mean = 0.0
        else:
            std_dev = 0.5
            mean = 0.5

        image_transforms = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize([mean], [std_dev]),
            ]
        )

        example["image"] = image_transforms(train_item["image"])

        if random.random() > self.conditional_dropout:
            example["tokens"] = self.tokenizer(
                train_item["caption"],
                truncation=True,
                padding="max_length",
                max_length=self.tokenizer.model_max_length,
            ).input_ids
        else:
            example["tokens"] = self.tokenizer(
                " ",
                truncation=True,
                padding="max_length",
                max_length=self.tokenizer.model_max_length,
            ).input_ids
        example["tokens"] = torch.tensor(example["tokens"])
        example["caption"] = train_item["caption"]  # for sampling if needed
        example["runt_size"] = train_item["runt_size"]

        return example

    def __get_image_for_trainer(self, image_train_item: ImageTrainItem, debug_level=0):
        example = {}
        save = debug_level > 2

        image_train_tmp = image_train_item.hydrate(crop=False, save=save, crop_jitter=self.crop_jitter)

        example["image"] = image_train_tmp.image
        example["caption"] = image_train_tmp.caption
        example["runt_size"] = image_train_tmp.runt_size
        return example



"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
from functools import wraps
from typing import TypeVar

from torch import amp, dtype, float16, no_grad

T = TypeVar("T")


def autocast_cuda(func: T, dtype: dtype = float16) -> T:
    @wraps(func)
    def wrap(*args, **kwargs):
        return amp.autocast(device_type="cuda", dtype=dtype)(no_grad()(func))(*args, **kwargs)

    return wrap


def autocast_cpu(func: T, dtype: dtype = float16) -> T:
    @wraps(func)
    def wrap(*args, **kwargs):
        return amp.autocast(device_type="cpu", dtype=dtype)(no_grad()(func))(*args, **kwargs)

    return wrap



"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import hashlib
import os
from concurrent.futures import ThreadPoolExecutor
from typing import Literal, Union

import numpy as np
from PIL import Image
from tqdm import tqdm

from nataili import disable_progress
from nataili.cache import Cache
from nataili.clip.image import ImageEmbed
from nataili.model_manager.clip import ClipModelManager
from nataili.util.logger import logger
from nataili.util.normalized import normalized


class BulkImageEmbedder:
    def __init__(self):
        self.model_manager = None
        self.model_name = None
        self.image_embed = None
        self.cache_image = None
        self.executor = ThreadPoolExecutor(max_workers=4)

    def __call__(
        self,
        model_name: Literal["ViT-L/14", "ViT-H-14"] = "ViT-L/14",
        input_directory: str = "input",
    ):
        """
        :param model_name: Name of model to use
        :param input_directory: Directory to read input from
        """
        self.model_name = model_name
        self.model_manager = ClipModelManager()
        self.cache_image = Cache(self.model_name, cache_parentname="embeds", cache_subname="image")
        self.model_manager.load(self.model_name)
        self.image_embed = ImageEmbed(self.model_manager.loaded_models[self.model_name], self.cache_image)
        self._prepare_from_directory(input_directory=input_directory)

    def insert(self, image, input_directory):
        pil_image = Image.open(f"{input_directory}/{image}.webp").convert("RGB")
        pil_hash = hashlib.sha256(pil_image.tobytes()).hexdigest()
        # hash = hashlib.sha256(open(f"{input_directory}/{image}.webp", "rb").read()).hexdigest()
        self.cache_image.add_sqlite_row(file=image, pil_hash=pil_hash, hash=None)

    def _prepare_from_directory(self, input_directory: str):
        """
        :param input_directory: Directory to read input from
        """
        logger.info(f"Reading from {input_directory}")
        directory_list = self.cache_image.list_dir(input_directory)
        logger.info(f"Found {len(directory_list)} files. Filtering...")
        filtered_list = self.cache_image.filter_list(directory_list)
        logger.info(f"Found {len(filtered_list)} files to embed.")
        for image in tqdm(filtered_list, disable=disable_progress.active):
            self.insert(image, input_directory)



"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import os


def save_sample(
    image,
    filename,
    sample_path,
    extension="png",
    jpg_quality=95,
    webp_quality=95,
    webp_lossless=True,
    png_compression=9,
):
    path = os.path.join(sample_path, filename + "." + extension)
    if os.path.exists(path):
        return False
    if not os.path.exists(sample_path):
        os.makedirs(sample_path)
    if extension == "png":
        image.save(path, format="PNG", compress_level=png_compression)
    elif extension == "jpg":
        image.save(path, quality=jpg_quality, optimize=True)
    elif extension == "webp":
        image.save(path, quality=webp_quality, lossless=webp_lossless)
    else:
        return False
    if os.path.exists(path):
        return True
    else:
        return False



# Copyright (c) OpenMMLab. All rights reserved.
from ...dist_utils import master_only
from ..hook import HOOKS
from .base import LoggerHook


@HOOKS.register_module()
class WandbLoggerHook(LoggerHook):

    def __init__(self,
                 init_kwargs=None,
                 interval=10,
                 ignore_last=True,
                 reset_flag=False,
                 commit=True,
                 by_epoch=True,
                 with_step=True):
        super(WandbLoggerHook, self).__init__(interval, ignore_last,
                                              reset_flag, by_epoch)
        self.import_wandb()
        self.init_kwargs = init_kwargs
        self.commit = commit
        self.with_step = with_step

    def import_wandb(self):
        try:
            import wandb
        except ImportError:
            raise ImportError(
                'Please run "pip install wandb" to install wandb')
        self.wandb = wandb

    @master_only
    def before_run(self, runner):
        super(WandbLoggerHook, self).before_run(runner)
        if self.wandb is None:
            self.import_wandb()
        if self.init_kwargs:
            self.wandb.init(**self.init_kwargs)
        else:
            self.wandb.init()

    @master_only
    def log(self, runner):
        tags = self.get_loggable_tags(runner)
        if tags:
            if self.with_step:
                self.wandb.log(
                    tags, step=self.get_iter(runner), commit=self.commit)
            else:
                tags['global_step'] = self.get_iter(runner)
                self.wandb.log(tags, commit=self.commit)

    @master_only
    def after_run(self, runner):
        self.wandb.join()



"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import contextlib
import math
from collections import namedtuple

import numpy as np
import torch
from PIL import Image

from nataili.util.postprocessor import PostProcessor

Grid = namedtuple("Grid", ["tiles", "tile_w", "tile_h", "image_w", "image_h", "overlap"])


class esrgan(PostProcessor):
    def set_filename_append(self):
        """
        Set filename append
        """
        self.filename_append = "esrgan"

    def process(self, img, img_array, **kwargs):
        """
        Override process method from PostProcessor
        :param img: PIL Image
        :param img_array: numpy array
        :param kwargs: strength
        :return: PIL Image
        """
        with contextlib.redirect_stdout(None):
            try:
                output, _ = self.model["model"].enhance(img_array)
                output_array = np.array(output)
                output_image = Image.fromarray(output_array)
            except:
                output_image = self.esrgan_upscale(model=self.model["model"], input_img=img)
        return output_image

    def esrgan_upscale(self, model, input_img):
        def esrgan_enhance(model, img):
            img = np.array(img)
            img = img[:, :, ::-1]
            img = np.ascontiguousarray(np.transpose(img, (2, 0, 1))) / 255
            img = torch.from_numpy(img).float()
            img = img.unsqueeze(0).to(self.model["device"])
            with torch.no_grad():
                output = model(img)
            output = output.squeeze().float().cpu().clamp_(0, 1).numpy()
            output = 255.0 * np.moveaxis(output, 0, 2)
            output = output.astype(np.uint8)
            output = output[:, :, ::-1]
            return Image.fromarray(output, "RGB")

        def split_grid(image, tile_w=512, tile_h=512, overlap=64):
            w = image.width
            h = image.height

            non_overlap_width = tile_w - overlap
            non_overlap_height = tile_h - overlap

            cols = math.ceil((w - overlap) / non_overlap_width)
            rows = math.ceil((h - overlap) / non_overlap_height)

            dx = (w - tile_w) / (cols - 1) if cols > 1 else 0
            dy = (h - tile_h) / (rows - 1) if rows > 1 else 0

            grid = Grid([], tile_w, tile_h, w, h, overlap)
            for row in range(rows):
                row_images = []

                y = int(row * dy)

                if y + tile_h >= h:
                    y = h - tile_h

                for col in range(cols):
                    x = int(col * dx)

                    if x + tile_w >= w:
                        x = w - tile_w

                    tile = image.crop((x, y, x + tile_w, y + tile_h))

                    row_images.append([x, tile_w, tile])

                grid.tiles.append([y, tile_h, row_images])

            return grid

        def combine_grid(grid):
            def make_mask_image(r):
                r = r * 255 / grid.overlap
                r = r.astype(np.uint8)
                return Image.fromarray(r, "L")

            mask_w = make_mask_image(
                np.arange(grid.overlap, dtype=np.float32).reshape((1, grid.overlap)).repeat(grid.tile_h, axis=0)
            )
            mask_h = make_mask_image(
                np.arange(grid.overlap, dtype=np.float32).reshape((grid.overlap, 1)).repeat(grid.image_w, axis=1)
            )

            combined_image = Image.new("RGB", (grid.image_w, grid.image_h))
            for y, h, row in grid.tiles:
                combined_row = Image.new("RGB", (grid.image_w, h))
                for x, w, tile in row:
                    if x == 0:
                        combined_row.paste(tile, (0, 0))
                        continue

                    combined_row.paste(tile.crop((0, 0, grid.overlap, h)), (x, 0), mask=mask_w)
                    combined_row.paste(tile.crop((grid.overlap, 0, w, h)), (x + grid.overlap, 0))

                if y == 0:
                    combined_image.paste(combined_row, (0, 0))
                    continue

                combined_image.paste(combined_row.crop((0, 0, combined_row.width, grid.overlap)), (0, y), mask=mask_h)
                combined_image.paste(
                    combined_row.crop((0, grid.overlap, combined_row.width, h)), (0, y + grid.overlap)
                )

            return combined_image

        grid = split_grid(image=input_img, tile_w=512, tile_h=512, overlap=64)
        newtiles = []
        scale_factor = 1

        for y, h, row in grid.tiles:
            newrow = []
            for tiledata in row:
                x, w, tile = tiledata

                output = esrgan_enhance(model=model, img=tile)
                scale_factor = output.width // tile.width

                newrow.append([x * scale_factor, w * scale_factor, output])
            newtiles.append([y * scale_factor, h * scale_factor, newrow])

        newgrid = Grid(
            newtiles,
            grid.tile_w * scale_factor,
            grid.tile_h * scale_factor,
            grid.image_w * scale_factor,
            grid.image_h * scale_factor,
            grid.overlap * scale_factor,
        )
        output = combine_grid(grid=newgrid)
        return output



# Copyright (c) OpenMMLab. All rights reserved.
import math

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from ..utils import kaiming_init
from .registry import PLUGIN_LAYERS


@PLUGIN_LAYERS.register_module()
class GeneralizedAttention(nn.Module):
    """GeneralizedAttention module.

    See 'An Empirical Study of Spatial Attention Mechanisms in Deep Networks'
    (https://arxiv.org/abs/1711.07971) for details.

    Args:
        in_channels (int): Channels of the input feature map.
        spatial_range (int): The spatial range. -1 indicates no spatial range
            constraint. Default: -1.
        num_heads (int): The head number of empirical_attention module.
            Default: 9.
        position_embedding_dim (int): The position embedding dimension.
            Default: -1.
        position_magnitude (int): A multiplier acting on coord difference.
            Default: 1.
        kv_stride (int): The feature stride acting on key/value feature map.
            Default: 2.
        q_stride (int): The feature stride acting on query feature map.
            Default: 1.
        attention_type (str): A binary indicator string for indicating which
            items in generalized empirical_attention module are used.
            Default: '1111'.

            - '1000' indicates 'query and key content' (appr - appr) item,
            - '0100' indicates 'query content and relative position'
              (appr - position) item,
            - '0010' indicates 'key content only' (bias - appr) item,
            - '0001' indicates 'relative position only' (bias - position) item.
    """

    _abbr_ = 'gen_attention_block'

    def __init__(self,
                 in_channels,
                 spatial_range=-1,
                 num_heads=9,
                 position_embedding_dim=-1,
                 position_magnitude=1,
                 kv_stride=2,
                 q_stride=1,
                 attention_type='1111'):

        super(GeneralizedAttention, self).__init__()

        # hard range means local range for non-local operation
        self.position_embedding_dim = (
            position_embedding_dim
            if position_embedding_dim > 0 else in_channels)

        self.position_magnitude = position_magnitude
        self.num_heads = num_heads
        self.in_channels = in_channels
        self.spatial_range = spatial_range
        self.kv_stride = kv_stride
        self.q_stride = q_stride
        self.attention_type = [bool(int(_)) for _ in attention_type]
        self.qk_embed_dim = in_channels // num_heads
        out_c = self.qk_embed_dim * num_heads

        if self.attention_type[0] or self.attention_type[1]:
            self.query_conv = nn.Conv2d(
                in_channels=in_channels,
                out_channels=out_c,
                kernel_size=1,
                bias=False)
            self.query_conv.kaiming_init = True

        if self.attention_type[0] or self.attention_type[2]:
            self.key_conv = nn.Conv2d(
                in_channels=in_channels,
                out_channels=out_c,
                kernel_size=1,
                bias=False)
            self.key_conv.kaiming_init = True

        self.v_dim = in_channels // num_heads
        self.value_conv = nn.Conv2d(
            in_channels=in_channels,
            out_channels=self.v_dim * num_heads,
            kernel_size=1,
            bias=False)
        self.value_conv.kaiming_init = True

        if self.attention_type[1] or self.attention_type[3]:
            self.appr_geom_fc_x = nn.Linear(
                self.position_embedding_dim // 2, out_c, bias=False)
            self.appr_geom_fc_x.kaiming_init = True

            self.appr_geom_fc_y = nn.Linear(
                self.position_embedding_dim // 2, out_c, bias=False)
            self.appr_geom_fc_y.kaiming_init = True

        if self.attention_type[2]:
            stdv = 1.0 / math.sqrt(self.qk_embed_dim * 2)
            appr_bias_value = -2 * stdv * torch.rand(out_c) + stdv
            self.appr_bias = nn.Parameter(appr_bias_value)

        if self.attention_type[3]:
            stdv = 1.0 / math.sqrt(self.qk_embed_dim * 2)
            geom_bias_value = -2 * stdv * torch.rand(out_c) + stdv
            self.geom_bias = nn.Parameter(geom_bias_value)

        self.proj_conv = nn.Conv2d(
            in_channels=self.v_dim * num_heads,
            out_channels=in_channels,
            kernel_size=1,
            bias=True)
        self.proj_conv.kaiming_init = True
        self.gamma = nn.Parameter(torch.zeros(1))

        if self.spatial_range >= 0:
            # only works when non local is after 3*3 conv
            if in_channels == 256:
                max_len = 84
            elif in_channels == 512:
                max_len = 42

            max_len_kv = int((max_len - 1.0) / self.kv_stride + 1)
            local_constraint_map = np.ones(
                (max_len, max_len, max_len_kv, max_len_kv), dtype=np.int)
            for iy in range(max_len):
                for ix in range(max_len):
                    local_constraint_map[
                        iy, ix,
                        max((iy - self.spatial_range) //
                            self.kv_stride, 0):min((iy + self.spatial_range +
                                                    1) // self.kv_stride +
                                                   1, max_len),
                        max((ix - self.spatial_range) //
                            self.kv_stride, 0):min((ix + self.spatial_range +
                                                    1) // self.kv_stride +
                                                   1, max_len)] = 0

            self.local_constraint_map = nn.Parameter(
                torch.from_numpy(local_constraint_map).byte(),
                requires_grad=False)

        if self.q_stride > 1:
            self.q_downsample = nn.AvgPool2d(
                kernel_size=1, stride=self.q_stride)
        else:
            self.q_downsample = None

        if self.kv_stride > 1:
            self.kv_downsample = nn.AvgPool2d(
                kernel_size=1, stride=self.kv_stride)
        else:
            self.kv_downsample = None

        self.init_weights()

    def get_position_embedding(self,
                               h,
                               w,
                               h_kv,
                               w_kv,
                               q_stride,
                               kv_stride,
                               device,
                               dtype,
                               feat_dim,
                               wave_length=1000):
        # the default type of Tensor is float32, leading to type mismatch
        # in fp16 mode. Cast it to support fp16 mode.
        h_idxs = torch.linspace(0, h - 1, h).to(device=device, dtype=dtype)
        h_idxs = h_idxs.view((h, 1)) * q_stride

        w_idxs = torch.linspace(0, w - 1, w).to(device=device, dtype=dtype)
        w_idxs = w_idxs.view((w, 1)) * q_stride

        h_kv_idxs = torch.linspace(0, h_kv - 1, h_kv).to(
            device=device, dtype=dtype)
        h_kv_idxs = h_kv_idxs.view((h_kv, 1)) * kv_stride

        w_kv_idxs = torch.linspace(0, w_kv - 1, w_kv).to(
            device=device, dtype=dtype)
        w_kv_idxs = w_kv_idxs.view((w_kv, 1)) * kv_stride

        # (h, h_kv, 1)
        h_diff = h_idxs.unsqueeze(1) - h_kv_idxs.unsqueeze(0)
        h_diff *= self.position_magnitude

        # (w, w_kv, 1)
        w_diff = w_idxs.unsqueeze(1) - w_kv_idxs.unsqueeze(0)
        w_diff *= self.position_magnitude

        feat_range = torch.arange(0, feat_dim / 4).to(
            device=device, dtype=dtype)

        dim_mat = torch.Tensor([wave_length]).to(device=device, dtype=dtype)
        dim_mat = dim_mat**((4. / feat_dim) * feat_range)
        dim_mat = dim_mat.view((1, 1, -1))

        embedding_x = torch.cat(
            ((w_diff / dim_mat).sin(), (w_diff / dim_mat).cos()), dim=2)

        embedding_y = torch.cat(
            ((h_diff / dim_mat).sin(), (h_diff / dim_mat).cos()), dim=2)

        return embedding_x, embedding_y

    def forward(self, x_input):
        num_heads = self.num_heads

        # use empirical_attention
        if self.q_downsample is not None:
            x_q = self.q_downsample(x_input)
        else:
            x_q = x_input
        n, _, h, w = x_q.shape

        if self.kv_downsample is not None:
            x_kv = self.kv_downsample(x_input)
        else:
            x_kv = x_input
        _, _, h_kv, w_kv = x_kv.shape

        if self.attention_type[0] or self.attention_type[1]:
            proj_query = self.query_conv(x_q).view(
                (n, num_heads, self.qk_embed_dim, h * w))
            proj_query = proj_query.permute(0, 1, 3, 2)

        if self.attention_type[0] or self.attention_type[2]:
            proj_key = self.key_conv(x_kv).view(
                (n, num_heads, self.qk_embed_dim, h_kv * w_kv))

        if self.attention_type[1] or self.attention_type[3]:
            position_embed_x, position_embed_y = self.get_position_embedding(
                h, w, h_kv, w_kv, self.q_stride, self.kv_stride,
                x_input.device, x_input.dtype, self.position_embedding_dim)
            # (n, num_heads, w, w_kv, dim)
            position_feat_x = self.appr_geom_fc_x(position_embed_x).\
                view(1, w, w_kv, num_heads, self.qk_embed_dim).\
                permute(0, 3, 1, 2, 4).\
                repeat(n, 1, 1, 1, 1)

            # (n, num_heads, h, h_kv, dim)
            position_feat_y = self.appr_geom_fc_y(position_embed_y).\
                view(1, h, h_kv, num_heads, self.qk_embed_dim).\
                permute(0, 3, 1, 2, 4).\
                repeat(n, 1, 1, 1, 1)

            position_feat_x /= math.sqrt(2)
            position_feat_y /= math.sqrt(2)

        # accelerate for saliency only
        if (np.sum(self.attention_type) == 1) and self.attention_type[2]:
            appr_bias = self.appr_bias.\
                view(1, num_heads, 1, self.qk_embed_dim).\
                repeat(n, 1, 1, 1)

            energy = torch.matmul(appr_bias, proj_key).\
                view(n, num_heads, 1, h_kv * w_kv)

            h = 1
            w = 1
        else:
            # (n, num_heads, h*w, h_kv*w_kv), query before key, 540mb for
            if not self.attention_type[0]:
                energy = torch.zeros(
                    n,
                    num_heads,
                    h,
                    w,
                    h_kv,
                    w_kv,
                    dtype=x_input.dtype,
                    device=x_input.device)

            # attention_type[0]: appr - appr
            # attention_type[1]: appr - position
            # attention_type[2]: bias - appr
            # attention_type[3]: bias - position
            if self.attention_type[0] or self.attention_type[2]:
                if self.attention_type[0] and self.attention_type[2]:
                    appr_bias = self.appr_bias.\
                        view(1, num_heads, 1, self.qk_embed_dim)
                    energy = torch.matmul(proj_query + appr_bias, proj_key).\
                        view(n, num_heads, h, w, h_kv, w_kv)

                elif self.attention_type[0]:
                    energy = torch.matmul(proj_query, proj_key).\
                        view(n, num_heads, h, w, h_kv, w_kv)

                elif self.attention_type[2]:
                    appr_bias = self.appr_bias.\
                        view(1, num_heads, 1, self.qk_embed_dim).\
                        repeat(n, 1, 1, 1)

                    energy += torch.matmul(appr_bias, proj_key).\
                        view(n, num_heads, 1, 1, h_kv, w_kv)

            if self.attention_type[1] or self.attention_type[3]:
                if self.attention_type[1] and self.attention_type[3]:
                    geom_bias = self.geom_bias.\
                        view(1, num_heads, 1, self.qk_embed_dim)

                    proj_query_reshape = (proj_query + geom_bias).\
                        view(n, num_heads, h, w, self.qk_embed_dim)

                    energy_x = torch.matmul(
                        proj_query_reshape.permute(0, 1, 3, 2, 4),
                        position_feat_x.permute(0, 1, 2, 4, 3))
                    energy_x = energy_x.\
                        permute(0, 1, 3, 2, 4).unsqueeze(4)

                    energy_y = torch.matmul(
                        proj_query_reshape,
                        position_feat_y.permute(0, 1, 2, 4, 3))
                    energy_y = energy_y.unsqueeze(5)

                    energy += energy_x + energy_y

                elif self.attention_type[1]:
                    proj_query_reshape = proj_query.\
                        view(n, num_heads, h, w, self.qk_embed_dim)
                    proj_query_reshape = proj_query_reshape.\
                        permute(0, 1, 3, 2, 4)
                    position_feat_x_reshape = position_feat_x.\
                        permute(0, 1, 2, 4, 3)
                    position_feat_y_reshape = position_feat_y.\
                        permute(0, 1, 2, 4, 3)

                    energy_x = torch.matmul(proj_query_reshape,
                                            position_feat_x_reshape)
                    energy_x = energy_x.permute(0, 1, 3, 2, 4).unsqueeze(4)

                    energy_y = torch.matmul(proj_query_reshape,
                                            position_feat_y_reshape)
                    energy_y = energy_y.unsqueeze(5)

                    energy += energy_x + energy_y

                elif self.attention_type[3]:
                    geom_bias = self.geom_bias.\
                        view(1, num_heads, self.qk_embed_dim, 1).\
                        repeat(n, 1, 1, 1)

                    position_feat_x_reshape = position_feat_x.\
                        view(n, num_heads, w*w_kv, self.qk_embed_dim)

                    position_feat_y_reshape = position_feat_y.\
                        view(n, num_heads, h * h_kv, self.qk_embed_dim)

                    energy_x = torch.matmul(position_feat_x_reshape, geom_bias)
                    energy_x = energy_x.view(n, num_heads, 1, w, 1, w_kv)

                    energy_y = torch.matmul(position_feat_y_reshape, geom_bias)
                    energy_y = energy_y.view(n, num_heads, h, 1, h_kv, 1)

                    energy += energy_x + energy_y

            energy = energy.view(n, num_heads, h * w, h_kv * w_kv)

        if self.spatial_range >= 0:
            cur_local_constraint_map = \
                self.local_constraint_map[:h, :w, :h_kv, :w_kv].\
                contiguous().\
                view(1, 1, h*w, h_kv*w_kv)

            energy = energy.masked_fill_(cur_local_constraint_map,
                                         float('-inf'))

        attention = F.softmax(energy, 3)

        proj_value = self.value_conv(x_kv)
        proj_value_reshape = proj_value.\
            view((n, num_heads, self.v_dim, h_kv * w_kv)).\
            permute(0, 1, 3, 2)

        out = torch.matmul(attention, proj_value_reshape).\
            permute(0, 1, 3, 2).\
            contiguous().\
            view(n, self.v_dim * self.num_heads, h, w)

        out = self.proj_conv(out)

        # output is downsampled, upsample back to input size
        if self.q_downsample is not None:
            out = F.interpolate(
                out,
                size=x_input.shape[2:],
                mode='bilinear',
                align_corners=False)

        out = self.gamma * out + x_input
        return out

    def init_weights(self):
        for m in self.modules():
            if hasattr(m, 'kaiming_init') and m.kaiming_init:
                kaiming_init(
                    m,
                    mode='fan_in',
                    nonlinearity='leaky_relu',
                    bias=0,
                    distribution='uniform',
                    a=1)



# Copyright (c) OpenMMLab. All rights reserved.
from .activation import build_activation_layer
from .context_block import ContextBlock
from .conv import build_conv_layer
from .conv2d_adaptive_padding import Conv2dAdaptivePadding
from .conv_module import ConvModule
from .conv_ws import ConvAWS2d, ConvWS2d, conv_ws_2d
from .depthwise_separable_conv_module import DepthwiseSeparableConvModule
from .drop import Dropout, DropPath
from .generalized_attention import GeneralizedAttention
from .hsigmoid import HSigmoid
from .hswish import HSwish
from .non_local import NonLocal1d, NonLocal2d, NonLocal3d
from .norm import build_norm_layer, is_norm
from .padding import build_padding_layer
from .plugin import build_plugin_layer
from .registry import (ACTIVATION_LAYERS, CONV_LAYERS, NORM_LAYERS,
                       PADDING_LAYERS, PLUGIN_LAYERS, UPSAMPLE_LAYERS)
from .scale import Scale
from .swish import Swish
from .upsample import build_upsample_layer
from .wrappers import (Conv2d, Conv3d, ConvTranspose2d, ConvTranspose3d,
                       Linear, MaxPool2d, MaxPool3d)

__all__ = [
    'ConvModule', 'build_activation_layer', 'build_conv_layer',
    'build_norm_layer', 'build_padding_layer', 'build_upsample_layer',
    'build_plugin_layer', 'is_norm', 'HSigmoid', 'HSwish', 'NonLocal1d',
    'NonLocal2d', 'NonLocal3d', 'ContextBlock', 'GeneralizedAttention',
    'ACTIVATION_LAYERS', 'CONV_LAYERS', 'NORM_LAYERS', 'PADDING_LAYERS',
    'UPSAMPLE_LAYERS', 'PLUGIN_LAYERS', 'Scale', 'ConvAWS2d', 'ConvWS2d',
    'conv_ws_2d', 'DepthwiseSeparableConvModule', 'Swish', 'Linear',
    'Conv2dAdaptivePadding', 'Conv2d', 'ConvTranspose2d', 'MaxPool2d',
    'ConvTranspose3d', 'MaxPool3d', 'Conv3d', 'Dropout', 'DropPath'
]



from uuid import uuid4

from nataili.model_manager.compvis import CompVisModelManager
from nataili.stable_diffusion.compvis import CompVis
from nataili.util.logger import logger

mm = CompVisModelManager()
base_dir = f"./test_output/{str(uuid4())}"
logger.info(f"Output dir: {base_dir}")
prompt = (
    "Headshot of cybernetic female character, cybernetic implants, solid background color,"
    "digital art, illustration, smooth color, cinematic moody lighting, cyberpunk, body modification,"
    "wenjun lin, studio ghibli, pixiv, artgerm, greg rutkowski, ilya kuvshinov"
)

for model in mm.models:
    output_dir = f"{base_dir}/{model}"
    logger.init(f"{model}", status="Loading")
    mm.load(model)
    logger.info(f"Running inference on {model}")
    compvis = CompVis(model=mm.loaded_models[model], output_dir=output_dir, model_name=model, disable_voodoo=True)
    compvis.generate(prompt)
    del mm.loaded_models[model]



import os
import random
import subprocess
import time
from argparse import ArgumentParser
from concurrent.futures import ThreadPoolExecutor

import boto3
import numpy as np
import webdataset as wds
from tqdm import tqdm

from nataili.cache import Cache
from nataili.clip.image import ImageEmbed
from nataili.model_manager.clip import ClipModelManager
from nataili.util.logger import logger

parser = ArgumentParser()
parser.add_argument("--start", type=int, help="start index", required=True)
parser.add_argument("--end", type=int, help="end index", required=True)
parser.add_argument("--zfill", type=int, default=6, help="zfill for index")
parser.add_argument("--batch_size", type=int, default=512)
parser.add_argument("--threads", type=int, default=4)
parser.add_argument("--endpoint", type=str, default=None, help="alternative endpoint for s3 i.e. R2")
parser.add_argument("--access_key", type=str, default=None)
parser.add_argument("--secret_key", type=str, default=None)
parser.add_argument("--bucket", type=str, required=True, help="s3://{bucket}")
parser.add_argument("--input_path", type=str, default=None, help="s3://{bucket}/{input_path}/{input}_{start..end}.tar")
parser.add_argument(
    "--output_path", type=str, default=None, help="s3://{bucket}/{output_path}/{output}_{start..end}.tar"
)
parser.add_argument("--input", type=str, required=True, help="s3://{bucket}/{input}_{start..end}.tar")
parser.add_argument("--output", type=str, required=True, help="s3://{bucket}/{output}_{start..end}.tar")
parser.add_argument("--gpu_id", type=int, default=0, help="gpu id to use")
parser.add_argument("--model", type=str, default="ViT-H-14", help="model to use")
parser.add_argument("--shard_size", type=int, default=10000, help="number of images per shard")
args = parser.parse_args()

executor = ThreadPoolExecutor(max_workers=args.threads, thread_name_prefix="SaveThread")
mm = ClipModelManager()
try:
    mm.load(args.model, args.gpu_id)
except:
    logger.error(f"Failed to load model {args.model}")
    exit(1)

s3_client = boto3.resource(
    "s3",
    endpoint_url=args.endpoint,
    aws_access_key_id=args.access_key if args.access_key else os.environ.get("AWS_ACCESS_KEY_ID"),
    aws_secret_access_key=args.secret_key if args.secret_key else os.environ.get("AWS_SECRET_ACCESS_KEY"),
)
bucket = s3_client.Bucket(args.bucket)


def embed_shard(shard_number: int, batch_size: int = 512):
    shard = f"{shard_number:0{args.zfill}d}"
    tar_file = f"{args.output}_{shard}.tar"
    try:
        bucket.Object(f"{args.output_path}/{tar_file}").last_modified
        logger.info(f"Shard {args.output}/{tar_file} already exists")
        return
    except:
        logger.info(f"Shard {args.output}/{tar_file} does not exist")
        pass
    if args.endpoint:
        url = f"pipe: aws s3 cp --endpoint-url {args.endpoint}"
    else:
        url = f"pipe: aws s3 cp"
    if args.input_path:
        url += f" s3://{args.bucket}/{args.input_path}/{args.input}_{shard}.tar -"
    else:
        url += f" s3://{args.bucket}/{args.input}_{shard}.tar -"
    dataset = wds.WebDataset(url).shuffle(100).decode("pil")
    cache_name = f"{args.input}_{shard}"
    cache = Cache(cache_name)
    db_files = cache.get_all()
    db_files = {file for file in db_files}
    logger.info(f"Found {len(db_files)} files in cache")
    batch = []
    progress_bar = tqdm(total=args.shard_size, unit="image", disable=False, desc=f"Shard {shard}")
    if len(db_files) != args.shard_size:
        for file in dataset:
            if file["__key__"] in db_files:
                progress_bar.total -= 1
                progress_bar.refresh()
                continue
            image = {}
            image["pil_image"] = file["webp"]
            image["filename"] = file["__key__"]
            batch.append(image)
            remaining = progress_bar.total - progress_bar.n
            if len(batch) == batch_size or remaining <= batch_size:
                image_embed = ImageEmbed(mm.loaded_models[args.model], cache)
                image_embed._batch(batch)
                del batch
                batch = []
                update_size = min(batch_size, remaining)
                progress_bar.update(update_size)
                progress_bar.refresh()
        logger.info(f"Finished embedding shard {shard}")
        logger.info(f"Waiting for cache to finish writing")
        time.sleep(10)
    export = cache.get_all_export()
    logger.info(f"Exporting shard {shard}")
    sink = wds.TarWriter(tar_file)
    export = sorted(export.items(), key=lambda x: x[0])
    export = {file: pil_hash for file, pil_hash in export}
    logger.info(f"Found {len(export)} files in cache")
    for file in export:
        pil_hash = export[file]
        embed_file = os.path.join(cache.cache_dir, pil_hash + ".npy")
        embed = np.load(embed_file)
        wds_data = {"__key__": file, "npy": embed}
        sink.write(wds_data)
    sink.close()
    logger.info(f"Finished exporting shard {shard} to {tar_file}")
    logger.info(f"Uploading shard {shard} to s3")
    #TODO: use boto3 instead of subprocess
    if args.endpoint:
        command = f"aws s3 cp --endpoint-url {args.endpoint}"
    else:
        command = f"aws s3 cp"
    if args.output_path:
        command += f" {tar_file} s3://{args.bucket}/{args.output_path}/{tar_file}"
    else:
        command += f" {tar_file} s3://{args.bucket}/{tar_file}"
    process = subprocess.check_output(command, shell=True)
    logger.info(f"Finished uploading shard {shard} to s3")


if __name__ == "__main__":
    total = args.end - args.start + 1
    logger.info(f"Embedding {total} shards")
    for shard in range(args.start, args.end + 1):
        if args.threads == 1:
            embed_shard(shard, args.batch_size)
        else:
            executor.submit(embed_shard, shard, args.batch_size)
        time.sleep(random.randint(1, 5))
    if args.threads > 1:
        executor.shutdown(wait=True)



import torch
import torch.nn.functional as F
from models.med import BertConfig, BertModel
from torch import nn
from transformers import BertTokenizer

from models.blip import create_vit, init_tokenizer, load_checkpoint


class BLIP_Retrieval(nn.Module):
    def __init__(self,                 
                 med_config = 'configs/med_config.json',  
                 image_size = 384,
                 vit = 'base',
                 vit_grad_ckpt = False,
                 vit_ckpt_layer = 0,                      
                 embed_dim = 256,     
                 queue_size = 57600,
                 momentum = 0.995,
                 negative_all_rank = False,
                 ):
        """
        Args:
            med_config (str): path for the mixture of encoder-decoder model's configuration file
            image_size (int): input image size
            vit (str): model size of vision transformer
        """               
        super().__init__()
        
        self.visual_encoder, vision_width = create_vit(vit,image_size, vit_grad_ckpt, vit_ckpt_layer)
        self.tokenizer = init_tokenizer()   
        med_config = BertConfig.from_json_file(med_config)
        med_config.encoder_width = vision_width
        self.text_encoder = BertModel(config=med_config, add_pooling_layer=False)          

        text_width = self.text_encoder.config.hidden_size
        
        self.vision_proj = nn.Linear(vision_width, embed_dim)
        self.text_proj = nn.Linear(text_width, embed_dim)

        self.itm_head = nn.Linear(text_width, 2) 
        
        # create momentum encoders  
        self.visual_encoder_m, vision_width = create_vit(vit,image_size)              
        self.vision_proj_m = nn.Linear(vision_width, embed_dim)
        self.text_encoder_m = BertModel(config=med_config, add_pooling_layer=False)    
        self.text_proj_m = nn.Linear(text_width, embed_dim)
        
        self.model_pairs = [[self.visual_encoder,self.visual_encoder_m],
                            [self.vision_proj,self.vision_proj_m],
                            [self.text_encoder,self.text_encoder_m],
                            [self.text_proj,self.text_proj_m],
                           ]       
        self.copy_params()

        # create the queue
        self.register_buffer("image_queue", torch.randn(embed_dim, queue_size))
        self.register_buffer("text_queue", torch.randn(embed_dim, queue_size))
        self.register_buffer("idx_queue", torch.full((1,queue_size),-100))
        self.register_buffer("ptr_queue", torch.zeros(1, dtype=torch.long))  

        self.image_queue = nn.functional.normalize(self.image_queue, dim=0)
        self.text_queue = nn.functional.normalize(self.text_queue, dim=0)
        
        self.queue_size = queue_size
        self.momentum = momentum
        self.temp = nn.Parameter(0.07*torch.ones([]))   
        
        self.negative_all_rank = negative_all_rank
        
        
    def forward(self, image, caption, alpha, idx):
        with torch.no_grad():
            self.temp.clamp_(0.001,0.5)
        
        image_embeds = self.visual_encoder(image) 
        image_atts = torch.ones(image_embeds.size()[:-1],dtype=torch.long).to(image.device)        
        image_feat = F.normalize(self.vision_proj(image_embeds[:,0,:]),dim=-1)    
        
        text = self.tokenizer(caption, padding='max_length', truncation=True, max_length=35, 
                              return_tensors="pt").to(image.device) 
        
        text_output = self.text_encoder(text.input_ids, attention_mask = text.attention_mask,                      
                                        return_dict = True, mode = 'text')            
        text_feat = F.normalize(self.text_proj(text_output.last_hidden_state[:,0,:]),dim=-1)        
        
        ###============== Image-text Contrastive Learning ===================###
        idx = idx.view(-1,1)
        idx_all = torch.cat([idx.t(), self.idx_queue.clone().detach()],dim=1)  
        pos_idx = torch.eq(idx, idx_all).float()       
        sim_targets = pos_idx / pos_idx.sum(1,keepdim=True)   
        
        # get momentum features
        with torch.no_grad():
            self._momentum_update()
            image_embeds_m = self.visual_encoder_m(image) 
            image_feat_m = F.normalize(self.vision_proj_m(image_embeds_m[:,0,:]),dim=-1)  
            image_feat_m_all = torch.cat([image_feat_m.t(),self.image_queue.clone().detach()],dim=1)                   
            
            text_output_m = self.text_encoder_m(text.input_ids, attention_mask = text.attention_mask,                      
                                                return_dict = True, mode = 'text')    
            text_feat_m = F.normalize(self.text_proj_m(text_output_m.last_hidden_state[:,0,:]),dim=-1) 
            text_feat_m_all = torch.cat([text_feat_m.t(),self.text_queue.clone().detach()],dim=1)

            sim_i2t_m = image_feat_m @ text_feat_m_all / self.temp  
            sim_t2i_m = text_feat_m @ image_feat_m_all / self.temp   

            sim_i2t_targets = alpha * F.softmax(sim_i2t_m, dim=1) + (1 - alpha) * sim_targets
            sim_t2i_targets = alpha * F.softmax(sim_t2i_m, dim=1) + (1 - alpha) * sim_targets        

        sim_i2t = image_feat @ text_feat_m_all / self.temp 
        sim_t2i = text_feat @ image_feat_m_all / self.temp 
                             
        loss_i2t = -torch.sum(F.log_softmax(sim_i2t, dim=1)*sim_i2t_targets,dim=1).mean()
        loss_t2i = -torch.sum(F.log_softmax(sim_t2i, dim=1)*sim_t2i_targets,dim=1).mean() 

        loss_ita = (loss_i2t+loss_t2i)/2
        
        idxs = concat_all_gather(idx)
        self._dequeue_and_enqueue(image_feat_m, text_feat_m, idxs)        

        ###============== Image-text Matching ===================###
        encoder_input_ids = text.input_ids.clone()
        encoder_input_ids[:,0] = self.tokenizer.enc_token_id

        # forward the positve image-text pair
        bs = image.size(0)
        output_pos = self.text_encoder(encoder_input_ids,
                                       attention_mask = text.attention_mask,
                                       encoder_hidden_states = image_embeds,
                                       encoder_attention_mask = image_atts,      
                                       return_dict = True,
                                      )  
        
        
        if self.negative_all_rank:    
            # compute sample similarity
            with torch.no_grad():                
                mask = torch.eq(idx, idxs.t())

                image_feat_world = concat_all_gather(image_feat)
                text_feat_world = concat_all_gather(text_feat)

                sim_i2t = image_feat @ text_feat_world.t() / self.temp 
                sim_t2i = text_feat @ image_feat_world.t() / self.temp 

                weights_i2t = F.softmax(sim_i2t,dim=1)
                weights_i2t.masked_fill_(mask, 0)            

                weights_t2i = F.softmax(sim_t2i,dim=1)
                weights_t2i.masked_fill_(mask, 0)     

            image_embeds_world = all_gather_with_grad(image_embeds) 

            # select a negative image (from all ranks) for each text
            image_embeds_neg = []    
            for b in range(bs):
                neg_idx = torch.multinomial(weights_t2i[b], 1).item()
                image_embeds_neg.append(image_embeds_world[neg_idx])
            image_embeds_neg = torch.stack(image_embeds_neg,dim=0)   

            # select a negative text (from all ranks) for each image
            input_ids_world = concat_all_gather(encoder_input_ids)
            att_mask_world = concat_all_gather(text.attention_mask)        

            text_ids_neg = []
            text_atts_neg = []
            for b in range(bs):
                neg_idx = torch.multinomial(weights_i2t[b], 1).item()
                text_ids_neg.append(input_ids_world[neg_idx])
                text_atts_neg.append(att_mask_world[neg_idx])
                
        else:
            with torch.no_grad():                
                mask = torch.eq(idx, idx.t())
                
                sim_i2t = image_feat @ text_feat.t() / self.temp 
                sim_t2i = text_feat @ image_feat.t() / self.temp 

                weights_i2t = F.softmax(sim_i2t,dim=1)
                weights_i2t.masked_fill_(mask, 0)            

                weights_t2i = F.softmax(sim_t2i,dim=1)
                weights_t2i.masked_fill_(mask, 0)     

            # select a negative image (from same rank) for each text
            image_embeds_neg = []    
            for b in range(bs):
                neg_idx = torch.multinomial(weights_t2i[b], 1).item()
                image_embeds_neg.append(image_embeds[neg_idx])
            image_embeds_neg = torch.stack(image_embeds_neg,dim=0)   

            # select a negative text (from same rank) for each image    
            text_ids_neg = []
            text_atts_neg = []
            for b in range(bs):
                neg_idx = torch.multinomial(weights_i2t[b], 1).item()
                text_ids_neg.append(encoder_input_ids[neg_idx])
                text_atts_neg.append(text.attention_mask[neg_idx])            
            
        text_ids_neg = torch.stack(text_ids_neg,dim=0)   
        text_atts_neg = torch.stack(text_atts_neg,dim=0)      

        text_ids_all = torch.cat([encoder_input_ids, text_ids_neg],dim=0)     
        text_atts_all = torch.cat([text.attention_mask, text_atts_neg],dim=0)     

        image_embeds_all = torch.cat([image_embeds_neg,image_embeds],dim=0)
        image_atts_all = torch.cat([image_atts,image_atts],dim=0)

        output_neg = self.text_encoder(text_ids_all,
                                       attention_mask = text_atts_all,
                                       encoder_hidden_states = image_embeds_all,
                                       encoder_attention_mask = image_atts_all,      
                                       return_dict = True,
                                      )                         
          

        vl_embeddings = torch.cat([output_pos.last_hidden_state[:,0,:], output_neg.last_hidden_state[:,0,:]],dim=0)
        vl_output = self.itm_head(vl_embeddings)            

        itm_labels = torch.cat([torch.ones(bs,dtype=torch.long),torch.zeros(2*bs,dtype=torch.long)],
                               dim=0).to(image.device)
        loss_itm = F.cross_entropy(vl_output, itm_labels)     

        return loss_ita, loss_itm 
 

    @torch.no_grad()    
    def copy_params(self):
        for model_pair in self.model_pairs:           
            for param, param_m in zip(model_pair[0].parameters(), model_pair[1].parameters()):
                param_m.data.copy_(param.data)  # initialize
                param_m.requires_grad = False  # not update by gradient    

            
    @torch.no_grad()        
    def _momentum_update(self):
        for model_pair in self.model_pairs:           
            for param, param_m in zip(model_pair[0].parameters(), model_pair[1].parameters()):
                param_m.data = param_m.data * self.momentum + param.data * (1. - self.momentum)
                
                
    @torch.no_grad()
    def _dequeue_and_enqueue(self, image_feat, text_feat, idxs):
        # gather keys before updating queue
        image_feats = concat_all_gather(image_feat)
        text_feats = concat_all_gather(text_feat)
        

        batch_size = image_feats.shape[0]

        ptr = int(self.ptr_queue)
        assert self.queue_size % batch_size == 0  # for simplicity

        # replace the keys at ptr (dequeue and enqueue)
        self.image_queue[:, ptr:ptr + batch_size] = image_feats.T
        self.text_queue[:, ptr:ptr + batch_size] = text_feats.T
        self.idx_queue[:, ptr:ptr + batch_size] = idxs.T
        ptr = (ptr + batch_size) % self.queue_size # move pointer

        self.ptr_queue[0] = ptr  


def blip_retrieval(pretrained='',**kwargs):
    model = BLIP_Retrieval(**kwargs)
    if pretrained:
        model,msg = load_checkpoint(model,pretrained)
        print("missing keys:")
        print(msg.missing_keys)
    return model 


@torch.no_grad()
def concat_all_gather(tensor):
    """
    Performs all_gather operation on the provided tensors.
    *** Warning ***: torch.distributed.all_gather has no gradient.
    """
    tensors_gather = [torch.ones_like(tensor)
        for _ in range(torch.distributed.get_world_size())]
    torch.distributed.all_gather(tensors_gather, tensor, async_op=False)

    output = torch.cat(tensors_gather, dim=0)
    return output      


class GatherLayer(torch.autograd.Function):
    """
    Gather tensors from all workers with support for backward propagation:
    This implementation does not cut the gradients as torch.distributed.all_gather does.
    """

    @staticmethod
    def forward(ctx, x):
        output = [torch.zeros_like(x) for _ in range(torch.distributed.get_world_size())]
        torch.distributed.all_gather(output, x)
        return tuple(output)

    @staticmethod
    def backward(ctx, *grads):
        all_gradients = torch.stack(grads)
        torch.distributed.all_reduce(all_gradients)
        return all_gradients[torch.distributed.get_rank()]


def all_gather_with_grad(tensors):
    """
    Performs all_gather operation on the provided tensors.
    Graph remains connected for backward grad computation.
    """
    # Queue the gathered tensors
    world_size = torch.distributed.get_world_size()
    # There is no need for reduction in the single-proc case
    if world_size == 1:
        return tensors

    tensor_all = GatherLayer.apply(tensors)

    return torch.cat(tensor_all, dim=0)



import logging
import warnings
from abc import ABCMeta, abstractmethod
from collections import OrderedDict

import annotator.uniformer.mmcv as mmcv
import numpy as np
import torch
import torch.distributed as dist
import torch.nn as nn
from annotator.uniformer.mmcv.runner import auto_fp16


class BaseSegmentor(nn.Module):
    """Base class for segmentors."""

    __metaclass__ = ABCMeta

    def __init__(self):
        super(BaseSegmentor, self).__init__()
        self.fp16_enabled = False

    @property
    def with_neck(self):
        """bool: whether the segmentor has neck"""
        return hasattr(self, 'neck') and self.neck is not None

    @property
    def with_auxiliary_head(self):
        """bool: whether the segmentor has auxiliary head"""
        return hasattr(self,
                       'auxiliary_head') and self.auxiliary_head is not None

    @property
    def with_decode_head(self):
        """bool: whether the segmentor has decode head"""
        return hasattr(self, 'decode_head') and self.decode_head is not None

    @abstractmethod
    def extract_feat(self, imgs):
        """Placeholder for extract features from images."""
        pass

    @abstractmethod
    def encode_decode(self, img, img_metas):
        """Placeholder for encode images with backbone and decode into a
        semantic segmentation map of the same size as input."""
        pass

    @abstractmethod
    def forward_train(self, imgs, img_metas, **kwargs):
        """Placeholder for Forward function for training."""
        pass

    @abstractmethod
    def simple_test(self, img, img_meta, **kwargs):
        """Placeholder for single image test."""
        pass

    @abstractmethod
    def aug_test(self, imgs, img_metas, **kwargs):
        """Placeholder for augmentation test."""
        pass

    def init_weights(self, pretrained=None):
        """Initialize the weights in segmentor.

        Args:
            pretrained (str, optional): Path to pre-trained weights.
                Defaults to None.
        """
        if pretrained is not None:
            logger = logging.getLogger()
            logger.info(f'load model from: {pretrained}')

    def forward_test(self, imgs, img_metas, **kwargs):
        """
        Args:
            imgs (List[Tensor]): the outer list indicates test-time
                augmentations and inner Tensor should have a shape NxCxHxW,
                which contains all images in the batch.
            img_metas (List[List[dict]]): the outer list indicates test-time
                augs (multiscale, flip, etc.) and the inner list indicates
                images in a batch.
        """
        for var, name in [(imgs, 'imgs'), (img_metas, 'img_metas')]:
            if not isinstance(var, list):
                raise TypeError(f'{name} must be a list, but got '
                                f'{type(var)}')

        num_augs = len(imgs)
        if num_augs != len(img_metas):
            raise ValueError(f'num of augmentations ({len(imgs)}) != '
                             f'num of image meta ({len(img_metas)})')
        # all images in the same aug batch all of the same ori_shape and pad
        # shape
        for img_meta in img_metas:
            ori_shapes = [_['ori_shape'] for _ in img_meta]
            assert all(shape == ori_shapes[0] for shape in ori_shapes)
            img_shapes = [_['img_shape'] for _ in img_meta]
            assert all(shape == img_shapes[0] for shape in img_shapes)
            pad_shapes = [_['pad_shape'] for _ in img_meta]
            assert all(shape == pad_shapes[0] for shape in pad_shapes)

        if num_augs == 1:
            return self.simple_test(imgs[0], img_metas[0], **kwargs)
        else:
            return self.aug_test(imgs, img_metas, **kwargs)

    @auto_fp16(apply_to=('img', ))
    def forward(self, img, img_metas, return_loss=True, **kwargs):
        """Calls either :func:`forward_train` or :func:`forward_test` depending
        on whether ``return_loss`` is ``True``.

        Note this setting will change the expected inputs. When
        ``return_loss=True``, img and img_meta are single-nested (i.e. Tensor
        and List[dict]), and when ``resturn_loss=False``, img and img_meta
        should be double nested (i.e.  List[Tensor], List[List[dict]]), with
        the outer list indicating test time augmentations.
        """
        if return_loss:
            return self.forward_train(img, img_metas, **kwargs)
        else:
            return self.forward_test(img, img_metas, **kwargs)

    def train_step(self, data_batch, optimizer, **kwargs):
        """The iteration step during training.

        This method defines an iteration step during training, except for the
        back propagation and optimizer updating, which are done in an optimizer
        hook. Note that in some complicated cases or models, the whole process
        including back propagation and optimizer updating is also defined in
        this method, such as GAN.

        Args:
            data (dict): The output of dataloader.
            optimizer (:obj:`torch.optim.Optimizer` | dict): The optimizer of
                runner is passed to ``train_step()``. This argument is unused
                and reserved.

        Returns:
            dict: It should contain at least 3 keys: ``loss``, ``log_vars``,
                ``num_samples``.
                ``loss`` is a tensor for back propagation, which can be a
                weighted sum of multiple losses.
                ``log_vars`` contains all the variables to be sent to the
                logger.
                ``num_samples`` indicates the batch size (when the model is
                DDP, it means the batch size on each GPU), which is used for
                averaging the logs.
        """
        losses = self(**data_batch)
        loss, log_vars = self._parse_losses(losses)

        outputs = dict(
            loss=loss,
            log_vars=log_vars,
            num_samples=len(data_batch['img_metas']))

        return outputs

    def val_step(self, data_batch, **kwargs):
        """The iteration step during validation.

        This method shares the same signature as :func:`train_step`, but used
        during val epochs. Note that the evaluation after training epochs is
        not implemented with this method, but an evaluation hook.
        """
        output = self(**data_batch, **kwargs)
        return output

    @staticmethod
    def _parse_losses(losses):
        """Parse the raw outputs (losses) of the network.

        Args:
            losses (dict): Raw output of the network, which usually contain
                losses and other necessary information.

        Returns:
            tuple[Tensor, dict]: (loss, log_vars), loss is the loss tensor
                which may be a weighted sum of all losses, log_vars contains
                all the variables to be sent to the logger.
        """
        log_vars = OrderedDict()
        for loss_name, loss_value in losses.items():
            if isinstance(loss_value, torch.Tensor):
                log_vars[loss_name] = loss_value.mean()
            elif isinstance(loss_value, list):
                log_vars[loss_name] = sum(_loss.mean() for _loss in loss_value)
            else:
                raise TypeError(
                    f'{loss_name} is not a tensor or list of tensors')

        loss = sum(_value for _key, _value in log_vars.items()
                   if 'loss' in _key)

        log_vars['loss'] = loss
        for loss_name, loss_value in log_vars.items():
            # reduce loss when distributed training
            if dist.is_available() and dist.is_initialized():
                loss_value = loss_value.data.clone()
                dist.all_reduce(loss_value.div_(dist.get_world_size()))
            log_vars[loss_name] = loss_value.item()

        return loss, log_vars

    def show_result(self,
                    img,
                    result,
                    palette=None,
                    win_name='',
                    show=False,
                    wait_time=0,
                    out_file=None,
                    opacity=0.5):
        """Draw `result` over `img`.

        Args:
            img (str or Tensor): The image to be displayed.
            result (Tensor): The semantic segmentation results to draw over
                `img`.
            palette (list[list[int]]] | np.ndarray | None): The palette of
                segmentation map. If None is given, random palette will be
                generated. Default: None
            win_name (str): The window name.
            wait_time (int): Value of waitKey param.
                Default: 0.
            show (bool): Whether to show the image.
                Default: False.
            out_file (str or None): The filename to write the image.
                Default: None.
            opacity(float): Opacity of painted segmentation map.
                Default 0.5.
                Must be in (0, 1] range.
        Returns:
            img (Tensor): Only if not `show` or `out_file`
        """
        img = mmcv.imread(img)
        img = img.copy()
        seg = result[0]
        if palette is None:
            if self.PALETTE is None:
                palette = np.random.randint(
                    0, 255, size=(len(self.CLASSES), 3))
            else:
                palette = self.PALETTE
        palette = np.array(palette)
        assert palette.shape[0] == len(self.CLASSES)
        assert palette.shape[1] == 3
        assert len(palette.shape) == 2
        assert 0 < opacity <= 1.0
        color_seg = np.zeros((seg.shape[0], seg.shape[1], 3), dtype=np.uint8)
        for label, color in enumerate(palette):
            color_seg[seg == label, :] = color
        # convert to BGR
        color_seg = color_seg[..., ::-1]

        img = img * (1 - opacity) + color_seg * opacity
        img = img.astype(np.uint8)
        # if out_file specified, do not show image in window
        if out_file is not None:
            show = False

        if show:
            mmcv.imshow(img, win_name, wait_time)
        if out_file is not None:
            mmcv.imwrite(img, out_file)

        if not (show or out_file):
            warnings.warn('show==False and out_file is not specified, only '
                          'result image will be returned')
            return img



"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import contextlib
import copy
import glob
import os
import pickle
import shutil
import warnings
from typing import Dict, List, Tuple, TypeVar

import ray
import torch

from nataili import InvalidModelCacheException, enable_local_ray_temp, enable_ray_alternative
from nataili.aitemplate import Model
from nataili.util.logger import logger

warnings.filterwarnings("ignore")


T = TypeVar("T")


def get_ray_temp_dir():
    return os.path.abspath(os.environ.get("RAY_TEMP_DIR", "./ray"))


def get_model_cache_dir():
    return os.path.join(get_ray_temp_dir(), "model-cache")


def initialise_voodoo():
    if enable_local_ray_temp.active:
        ray_temp_dir = get_ray_temp_dir()
        session_dirs = glob.glob(os.path.join(ray_temp_dir, "session_*"))
        for adir in session_dirs:
            shutil.rmtree(adir, ignore_errors=True)
        os.makedirs(ray_temp_dir, exist_ok=True)
        ray.init(_temp_dir=ray_temp_dir)
        logger.init(f"Ray temp dir '{ray_temp_dir}'", status="Prepared")
    else:
        logger.init_warn("Ray temp dir'", status="OS Default")


def extract_tensors(m: torch.nn.Module) -> Tuple[torch.nn.Module, List[Dict]]:
    tensors = []
    for _, module in m.named_modules():
        params = {
            name: torch.clone(param).cpu().detach().numpy() for name, param in module.named_parameters(recurse=False)
        }
        buffers = {name: torch.clone(buf).cpu().detach().numpy() for name, buf in module.named_buffers(recurse=False)}
        tensors.append({"params": params, "buffers": buffers})

    m_copy = copy.deepcopy(m)
    for _, module in m_copy.named_modules():
        for name in [name for name, _ in module.named_parameters(recurse=False)] + [
            name for name, _ in module.named_buffers(recurse=False)
        ]:
            setattr(module, name, None)

    m_copy.train(False)
    return m_copy, tensors


def replace_tensors(m: torch.nn.Module, tensors: List[Dict], device="cuda"):
    modules = [module for _, module in m.named_modules()]
    for module, tensor_dict in zip(modules, tensors):
        for name, array in tensor_dict["params"].items():
            module.register_parameter(
                name,
                torch.nn.Parameter(torch.as_tensor(array, device=device), requires_grad=False),
            )
        for name, array in tensor_dict["buffers"].items():
            module.register_buffer(name, torch.as_tensor(array, device=device))


def get_model_cache_filename(model_filename):
    return os.path.join(get_model_cache_dir(), os.path.basename(model_filename)) + ".cache"


def have_model_cache(model_filename):
    cache_file = get_model_cache_filename(model_filename)
    if os.path.exists(cache_file):
        # We have a cache file but only consider it valid if it's up to date
        model_timestamp = os.path.getmtime(model_filename)
        cache_timestamp = os.path.getmtime(cache_file)
        if model_timestamp <= cache_timestamp:
            return True
    return False


@contextlib.contextmanager
def load_from_plasma(ref, device="cuda"):
    if not enable_ray_alternative.active:
        # Load object from ray object store
        skeleton, weights = ray.get(ref)
    else:
        try:
            # Load object from our persistent store
            with open(ref, "rb") as cache:
                skeleton, weights = pickle.load(cache)
        except (pickle.PickleError, EOFError):
            # Most likely corrupt cache file, remove the file
            try:
                os.remove(ref)
            except OSError:
                pass  # we tried
            raise InvalidModelCacheException(f"Model .cache file {ref} was corrupt. It has been removed.")
    replace_tensors(skeleton, weights, device=device)
    skeleton.eval().to(device, memory_format=torch.channels_last)
    yield skeleton
    torch.cuda.empty_cache()


def push_model_to_plasma(model: torch.nn.Module, filename="") -> ray.ObjectRef:
    if not enable_ray_alternative.active:
        # Store object in ray object store
        ref = ray.put(extract_tensors(model))
    else:
        # Store object directly on disk
        cachefile = get_model_cache_filename(filename)
        if have_model_cache(cachefile):
            return cachefile
        # Create cache directory if it doesn't already exist
        if not os.path.isdir(get_model_cache_dir()):
            os.makedirs(get_model_cache_dir(), exist_ok=True)
        # Serialise our object
        with open(cachefile, "wb") as cache:
            pickle.dump(extract_tensors(model), cache, protocol=pickle.HIGHEST_PROTOCOL)
        ref = cachefile

    return ref


@contextlib.contextmanager
def load_diffusers_pipeline_from_plasma(ref, device="cuda"):
    pipe, modules = ray.get(ref)
    for name, weights in modules.items():
        replace_tensors(getattr(pipe, name), weights, device=device)
    pipe.to(device)
    yield pipe
    torch.cuda.empty_cache()


def push_diffusers_pipeline_to_plasma(pipe) -> ray.ObjectRef:
    modules = {}
    components = pipe.components
    for name, component in components.items():
        if isinstance(component, torch.nn.Module):
            skeleton, weights = extract_tensors(component)
            setattr(pipe, name, skeleton)
            modules[name] = weights
    ref = ray.put((pipe, modules))
    return ref


def init_ait_module(
    model_name,
    workdir,
):
    mod = Model(os.path.join(workdir, model_name))
    return mod


def push_ait_module(module: Model) -> ray.ObjectRef:
    ref = ray.put(module)
    return ref


def load_ait_module(ref):
    ait_module = ray.get(ref)
    return ait_module



"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import random


def seed_to_int(s):
    if type(s) is int:
        return s
    if s is None or s == "":
        return random.randint(0, 2**32 - 1)

    if type(s) is list:
        seed_list = []
        for seed in s:
            if seed is None or seed == "":
                seed_list.append(random.randint(0, 2**32 - 1))
            else:
                seed_list = s

        return seed_list

    n = abs(int(s) if s.isdigit() else random.Random(s).randint(0, 2**32 - 1))
    while n >= 2**32:
        n = n >> 32
    return n



from nataili.train.dreambooth import DreamboothLoRA

trainer = DreamboothLoRA()

trainer.train(
    base_checkpoint= "runwayml/stable-diffusion-v1-5",
    data_root= "X:/lora/squishmallow",
    project_name= "squishmallow",
    output_dir= "X:/lora/output",
    max_train_steps= 1000,
)


"""
This file is part of nataili ("Homepage" = "https://github.com/db0/nataili").

Copyright 2022-2023 hlky. Copyright 2023 hlky and AI Horde Community
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
import hashlib
from concurrent.futures import ThreadPoolExecutor

import clip
import numpy as np
import torch

from nataili.cache import Cache
from nataili.util.cast import autocast_cuda
from nataili.util.logger import logger


class TextEmbed:
    def __init__(self, model, cache: Cache):
        """
        :param model: Loaded model from ModelManager
        :param cache: Cache object
        """
        self.model = model
        self.cache = cache
        self.executor = ThreadPoolExecutor(max_workers=1024, thread_name_prefix="SaveThread")

    @autocast_cuda
    def _batch(self, text_list: list):
        for text in text_list:
            if isinstance(text["prompt"], bytes):
                text["prompt"] = text["prompt"].decode("utf-8")
            text["hash"] = hashlib.sha256(text["prompt"]).hexdigest()
        text_tokens = [clip.tokenize(text["prompt"], truncate=True).to(self.model["device"]) for text in text_list]
        text_tokens = torch.cat(text_tokens, dim=0)
        with torch.no_grad():
            text_features = self.model["model"].encode_text(text_tokens).float()
        for text_embed_array, text in zip(text_features, text_list):
            future = self.executor.submit(self._save, text_embed_array, text["hash"])
            self.cache.add_sqlite_row(text["filename"], text["hash"], text["hash"])

    def _save(self, text_embed_array, text_hash):
        text_embed_array /= text_embed_array.norm(dim=-1, keepdim=True)
        np.save(f"{self.cache.cache_dir}/{text_hash}", text_embed_array.float().cpu().detach().numpy())

    @autocast_cuda
    def __call__(self, text: str, check_cache: bool = False, filename: str = None):
        """
        :param text: Text to embed
        :param check_cache: Check cache for text
        :param filename: Filename to save to cache
        If text is not in cache, embed it and save it to cache
        """
        if isinstance(text, bytes):
            text = text.decode("utf-8")
        text_hash = hashlib.sha256(text.encode("utf-8")).hexdigest()
        if check_cache:
            cached = self.cache.get(file_hash=text_hash)
            if cached:
                return cached
        text_tokens = clip.tokenize([text], truncate=True).to(self.model["device"])
        with torch.no_grad():
            text_features = self.model["model"].encode_text(text_tokens).float()
        text_features /= text_features.norm(dim=-1, keepdim=True)
        text_embed_array = text_features.cpu().detach().numpy()
        if filename:
            np.save(f"{self.cache.cache_dir}/{text_hash}", text_embed_array)
            self.cache.add_sqlite_row(filename, text_hash, text_hash)
        else:
            np.save(f"{self.cache.cache_dir}/{text_hash}", text_embed_array)
            self.cache.add_sqlite_row(text, text_hash, text_hash)



'''
 * Copyright (c) 2022, salesforce.com, inc.
 * All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 * For full license text, see LICENSE.txt file in the repo root or https://opensource.org/licenses/BSD-3-Clause
 * By Junnan Li
 * Based on timm code base
 * https://github.com/rwightman/pytorch-image-models/tree/master/timm
'''

from functools import partial

import torch
import torch.nn as nn
import torch.nn.functional as F
from fairscale.nn.checkpoint.checkpoint_activations import checkpoint_wrapper
from timm.models.helpers import adapt_input_conv, named_apply
from timm.models.layers import DropPath, trunc_normal_
from timm.models.registry import register_model
from timm.models.vision_transformer import PatchEmbed, _cfg


class Mlp(nn.Module):
    """ MLP as used in Vision Transformer, MLP-Mixer and related networks
    """
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        x = self.drop(x)
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0., proj_drop=0.):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        # NOTE scale factor was wrong in my original version, can set manually to be compat with prev weights
        self.scale = qk_scale or head_dim ** -0.5
        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)
        self.attn_gradients = None
        self.attention_map = None
        
    def save_attn_gradients(self, attn_gradients):
        self.attn_gradients = attn_gradients
        
    def get_attn_gradients(self):
        return self.attn_gradients
    
    def save_attention_map(self, attention_map):
        self.attention_map = attention_map
        
    def get_attention_map(self):
        return self.attention_map
    
    def forward(self, x, register_hook=False):
        B, N, C = x.shape
        qkv = self.qkv(x).reshape(B, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]   # make torchscript happy (cannot use tensor as tuple)

        attn = (q @ k.transpose(-2, -1)) * self.scale
        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)
                
        if register_hook:
            self.save_attention_map(attn)
            attn.register_hook(self.save_attn_gradients)        

        x = (attn @ v).transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)
        x = self.proj_drop(x)
        return x


class Block(nn.Module):

    def __init__(self, dim, num_heads, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop=0., attn_drop=0.,
                 drop_path=0., act_layer=nn.GELU, norm_layer=nn.LayerNorm, use_grad_checkpointing=False):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim, num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)
        # NOTE: drop path for stochastic depth, we shall see if this is better than dropout here
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

        if use_grad_checkpointing:
            self.attn = checkpoint_wrapper(self.attn)
            self.mlp = checkpoint_wrapper(self.mlp)

    def forward(self, x, register_hook=False):
        x = x + self.drop_path(self.attn(self.norm1(x), register_hook=register_hook))
        x = x + self.drop_path(self.mlp(self.norm2(x)))
        return x

    
class VisionTransformer(nn.Module):
    """ Vision Transformer
    A PyTorch impl of : `An Image is Worth 16x16 Words: Transformers for Image Recognition at Scale`  -
        https://arxiv.org/abs/2010.11929
    """
    def __init__(self, img_size=224, patch_size=16, in_chans=3, num_classes=1000, embed_dim=768, depth=12,
                 num_heads=12, mlp_ratio=4., qkv_bias=True, qk_scale=None, representation_size=None,
                 drop_rate=0., attn_drop_rate=0., drop_path_rate=0., norm_layer=None, 
                 use_grad_checkpointing=False, ckpt_layer=0):
        """
        Args:
            img_size (int, tuple): input image size
            patch_size (int, tuple): patch size
            in_chans (int): number of input channels
            num_classes (int): number of classes for classification head
            embed_dim (int): embedding dimension
            depth (int): depth of transformer
            num_heads (int): number of attention heads
            mlp_ratio (int): ratio of mlp hidden dim to embedding dim
            qkv_bias (bool): enable bias for qkv if True
            qk_scale (float): override default qk scale of head_dim ** -0.5 if set
            representation_size (Optional[int]): enable and set representation layer (pre-logits) to this value if set
            drop_rate (float): dropout rate
            attn_drop_rate (float): attention dropout rate
            drop_path_rate (float): stochastic depth rate
            norm_layer: (nn.Module): normalization layer
        """
        super().__init__()
        self.num_features = self.embed_dim = embed_dim  # num_features for consistency with other models
        norm_layer = norm_layer or partial(nn.LayerNorm, eps=1e-6)

        self.patch_embed = PatchEmbed(
            img_size=img_size, patch_size=patch_size, in_chans=in_chans, embed_dim=embed_dim)

        num_patches = self.patch_embed.num_patches

        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + 1, embed_dim))
        self.pos_drop = nn.Dropout(p=drop_rate)

        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # stochastic depth decay rule
        self.blocks = nn.ModuleList([
            Block(
                dim=embed_dim, num_heads=num_heads, mlp_ratio=mlp_ratio, qkv_bias=qkv_bias, qk_scale=qk_scale,
                drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[i], norm_layer=norm_layer,
                use_grad_checkpointing=(use_grad_checkpointing and i>=depth-ckpt_layer)
            )
            for i in range(depth)])
        self.norm = norm_layer(embed_dim)

        trunc_normal_(self.pos_embed, std=.02)
        trunc_normal_(self.cls_token, std=.02)
        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    @torch.jit.ignore
    def no_weight_decay(self):
        return {'pos_embed', 'cls_token'}

    def forward(self, x, register_blk=-1):
        B = x.shape[0]
        x = self.patch_embed(x)

        cls_tokens = self.cls_token.expand(B, -1, -1)  # stole cls_tokens impl from Phil Wang, thanks
        x = torch.cat((cls_tokens, x), dim=1)
  
        x = x + self.pos_embed[:,:x.size(1),:]
        x = self.pos_drop(x)

        for i,blk in enumerate(self.blocks):
            x = blk(x, register_blk==i)
        x = self.norm(x)
        
        return x

    @torch.jit.ignore()
    def load_pretrained(self, checkpoint_path, prefix=''):
        _load_weights(self, checkpoint_path, prefix)
        

@torch.no_grad()
def _load_weights(model: VisionTransformer, checkpoint_path: str, prefix: str = ''):
    """ Load weights from .npz checkpoints for official Google Brain Flax implementation
    """
    import numpy as np

    def _n2p(w, t=True):
        if w.ndim == 4 and w.shape[0] == w.shape[1] == w.shape[2] == 1:
            w = w.flatten()
        if t:
            if w.ndim == 4:
                w = w.transpose([3, 2, 0, 1])
            elif w.ndim == 3:
                w = w.transpose([2, 0, 1])
            elif w.ndim == 2:
                w = w.transpose([1, 0])
        return torch.from_numpy(w)

    w = np.load(checkpoint_path)
    if not prefix and 'opt/target/embedding/kernel' in w:
        prefix = 'opt/target/'

    if hasattr(model.patch_embed, 'backbone'):
        # hybrid
        backbone = model.patch_embed.backbone
        stem_only = not hasattr(backbone, 'stem')
        stem = backbone if stem_only else backbone.stem
        stem.conv.weight.copy_(adapt_input_conv(stem.conv.weight.shape[1], _n2p(w[f'{prefix}conv_root/kernel'])))
        stem.norm.weight.copy_(_n2p(w[f'{prefix}gn_root/scale']))
        stem.norm.bias.copy_(_n2p(w[f'{prefix}gn_root/bias']))
        if not stem_only:
            for i, stage in enumerate(backbone.stages):
                for j, block in enumerate(stage.blocks):
                    bp = f'{prefix}block{i + 1}/unit{j + 1}/'
                    for r in range(3):
                        getattr(block, f'conv{r + 1}').weight.copy_(_n2p(w[f'{bp}conv{r + 1}/kernel']))
                        getattr(block, f'norm{r + 1}').weight.copy_(_n2p(w[f'{bp}gn{r + 1}/scale']))
                        getattr(block, f'norm{r + 1}').bias.copy_(_n2p(w[f'{bp}gn{r + 1}/bias']))
                    if block.downsample is not None:
                        block.downsample.conv.weight.copy_(_n2p(w[f'{bp}conv_proj/kernel']))
                        block.downsample.norm.weight.copy_(_n2p(w[f'{bp}gn_proj/scale']))
                        block.downsample.norm.bias.copy_(_n2p(w[f'{bp}gn_proj/bias']))
        embed_conv_w = _n2p(w[f'{prefix}embedding/kernel'])
    else:
        embed_conv_w = adapt_input_conv(
            model.patch_embed.proj.weight.shape[1], _n2p(w[f'{prefix}embedding/kernel']))
    model.patch_embed.proj.weight.copy_(embed_conv_w)
    model.patch_embed.proj.bias.copy_(_n2p(w[f'{prefix}embedding/bias']))
    model.cls_token.copy_(_n2p(w[f'{prefix}cls'], t=False))
    pos_embed_w = _n2p(w[f'{prefix}Transformer/posembed_input/pos_embedding'], t=False)
    if pos_embed_w.shape != model.pos_embed.shape:
        pos_embed_w = resize_pos_embed(  # resize pos embedding when different size from pretrained weights
            pos_embed_w, model.pos_embed, getattr(model, 'num_tokens', 1), model.patch_embed.grid_size)
    model.pos_embed.copy_(pos_embed_w)
    model.norm.weight.copy_(_n2p(w[f'{prefix}Transformer/encoder_norm/scale']))
    model.norm.bias.copy_(_n2p(w[f'{prefix}Transformer/encoder_norm/bias']))
#     if isinstance(model.head, nn.Linear) and model.head.bias.shape[0] == w[f'{prefix}head/bias'].shape[-1]:
#         model.head.weight.copy_(_n2p(w[f'{prefix}head/kernel']))
#         model.head.bias.copy_(_n2p(w[f'{prefix}head/bias']))
#     if isinstance(getattr(model.pre_logits, 'fc', None), nn.Linear) and f'{prefix}pre_logits/bias' in w:
#         model.pre_logits.fc.weight.copy_(_n2p(w[f'{prefix}pre_logits/kernel']))
#         model.pre_logits.fc.bias.copy_(_n2p(w[f'{prefix}pre_logits/bias']))
    for i, block in enumerate(model.blocks.children()):
        block_prefix = f'{prefix}Transformer/encoderblock_{i}/'
        mha_prefix = block_prefix + 'MultiHeadDotProductAttention_1/'
        block.norm1.weight.copy_(_n2p(w[f'{block_prefix}LayerNorm_0/scale']))
        block.norm1.bias.copy_(_n2p(w[f'{block_prefix}LayerNorm_0/bias']))
        block.attn.qkv.weight.copy_(torch.cat([
            _n2p(w[f'{mha_prefix}{n}/kernel'], t=False).flatten(1).T for n in ('query', 'key', 'value')]))
        block.attn.qkv.bias.copy_(torch.cat([
            _n2p(w[f'{mha_prefix}{n}/bias'], t=False).reshape(-1) for n in ('query', 'key', 'value')]))
        block.attn.proj.weight.copy_(_n2p(w[f'{mha_prefix}out/kernel']).flatten(1))
        block.attn.proj.bias.copy_(_n2p(w[f'{mha_prefix}out/bias']))
        for r in range(2):
            getattr(block.mlp, f'fc{r + 1}').weight.copy_(_n2p(w[f'{block_prefix}MlpBlock_3/Dense_{r}/kernel']))
            getattr(block.mlp, f'fc{r + 1}').bias.copy_(_n2p(w[f'{block_prefix}MlpBlock_3/Dense_{r}/bias']))
        block.norm2.weight.copy_(_n2p(w[f'{block_prefix}LayerNorm_2/scale']))
        block.norm2.bias.copy_(_n2p(w[f'{block_prefix}LayerNorm_2/bias']))

            
def interpolate_pos_embed(pos_embed_checkpoint, visual_encoder):        
    # interpolate position embedding
    embedding_size = pos_embed_checkpoint.shape[-1]
    num_patches = visual_encoder.patch_embed.num_patches
    num_extra_tokens = visual_encoder.pos_embed.shape[-2] - num_patches
    # height (== width) for the checkpoint position embedding
    orig_size = int((pos_embed_checkpoint.shape[-2] - num_extra_tokens) ** 0.5)
    # height (== width) for the new position embedding
    new_size = int(num_patches ** 0.5)

    if orig_size!=new_size:
        # class_token and dist_token are kept unchanged
        extra_tokens = pos_embed_checkpoint[:, :num_extra_tokens]
        # only the position tokens are interpolated
        pos_tokens = pos_embed_checkpoint[:, num_extra_tokens:]
        pos_tokens = pos_tokens.reshape(-1, orig_size, orig_size, embedding_size).permute(0, 3, 1, 2)
        pos_tokens = torch.nn.functional.interpolate(
            pos_tokens, size=(new_size, new_size), mode='bicubic', align_corners=False)
        pos_tokens = pos_tokens.permute(0, 2, 3, 1).flatten(1, 2)
        new_pos_embed = torch.cat((extra_tokens, pos_tokens), dim=1)
        print('reshape position embedding from %d to %d'%(orig_size ** 2,new_size ** 2))
        
        return new_pos_embed    
    else:
        return pos_embed_checkpoint